import re
import sys
import warnings
from collections.abc import Iterable
from datetime import date, datetime
from typing import Any, TypeAlias, cast

from cognite.client.data_classes import Asset, Relationship
from cognite.client.data_classes.data_modeling import EdgeApply, MappedPropertyApply, NodeApply, NodeOrEdgeData, ViewId
from cognite.client.data_classes.data_modeling.views import SingleHopConnectionDefinitionApply, ViewApply
from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic._internal._model_construction import ModelMetaclass
from rdflib import Graph, URIRef
from typing_extensions import TypeAliasType

from cognite.neat.legacy.graph.loaders.core.rdf_to_assets import NeatMetadataKeys
from cognite.neat.legacy.graph.transformations.query_generator.sparql import build_construct_query, triples2dictionary
from cognite.neat.legacy.rules import exceptions
from cognite.neat.legacy.rules.analysis import define_class_asset_mapping, to_class_property_pairs
from cognite.neat.legacy.rules.exporters._rules2dms import DMSSchemaComponents
from cognite.neat.legacy.rules.exporters._validation import are_entity_names_dms_compliant
from cognite.neat.legacy.rules.models.rules import Property, Rules
from cognite.neat.legacy.rules.models.value_types import ValueTypeMapping
from cognite.neat.utils.utils import create_sha256_hash, generate_exception_report

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc

EdgeOneToOne: TypeAlias = TypeAliasType("EdgeOneToOne", str)  # type: ignore[valid-type]
EdgeOneToMany: TypeAlias = TypeAliasType("EdgeOneToMany", list[str])  # type: ignore[valid-type]


def default_model_configuration(
    external_id: str | None = None,
    name: str | None = None,
    description: str | None = None,
    space: str | None = None,
    version: str | None = None,
) -> ConfigDict:
    return ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="allow",
        json_schema_extra={
            "title": name,
            "description": description,
            "external_id": external_id,
            "name": name,
            "space": space,
            "version": version,
        },
    )


def default_class_methods():
    return [
        from_dict,
        from_graph,
        to_asset,
        to_relationship,
        to_node,
        to_edge,
        get_field_description,
        get_field_name,
    ]


def default_class_property_methods():
    return [model_name, model_external_id, model_description, attributes, edges_one_to_one, edges_one_to_many]


def rules_to_pydantic_models(
    rules: Rules,
    methods: list | None = None,
    add_extra_fields: bool = False,
) -> dict[str, ModelMetaclass]:
    """
    Generate pydantic models from rules.

    Args:
        rules: Rules to generate pydantic models from
        methods: List of methods to register for pydantic models,
                 by default None meaning defaulting to base neat methods.
        add_extra_fields: Flag indicating to add extra fields to pydantic models, by default False

    Returns:
        Dictionary containing pydantic models with class ids as key and pydantic model as value
        containing properties defined for the given class(es) in the rules.


    !!! note "Default NEAT methods"
        Default NEAT methods which are added to the generated pydantic models are:

        - `get_field_description`: Returns description of the field if one exists.
        - `get_field_name`: Returns name of the field if one exists.
        - `model_name`: Returns the name of the model if one exists.
        - `model_external_id`: Returns the external id of the model if one exists.
        - `model_description`: Returns the description of the model if one exists.
        - `from_graph`: Creates model instance from class instance stored in RDF graph.
        - `to_asset`: Creates CDF Asset instance from model instance.
        - `to_node`: Creates DMS node from model instance.
        - `to_edge`: Creates DMS edge from model instance.
        - `attributes`: Returns list of node attributes.
        - `edges_one_to_one`: Returns list of node edges one to one.
        - `edges_one_to_many`: Returns list of node edges one to many.


    !!! note "Limitations"
        Currently this will take only unique properties and those which column rule_type
        is set to rdfpath, hence only_rdfpath = True. This means that at the moment
        we do not support UNION, i.e. ability to handle multiple rdfpaths for the same
        property. This is needed option and should be added in the second version of the exporter.

    !!! note "Classes and Properties must be DMS Compliant"
        Rules must be DMS compliant, i.e. class and property ids must obey the following regex:
        r`(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)` and must not contain reserved keywords.

        Classes reserved keywords: `Query`, `Mutation`, `Subscription`, `String`, `Int32`,
        `Int64`, `Int`, `Float32`, `Float64`, `Float`,`Timestamp`, `JSONObject`, `Date`,
        `Numeric`, `Boolean`, `PageInfo`, `File`, `Sequence`, `TimeSeries`

        Properties reserved keywords: `space`, `externalId`, `createdTime`, `lastUpdatedTime`,
        `deletedTime`, `edge_id` `node_id`, `project_id`, `property_group`, `seq`, `tg_table_name`, `extensions`
    """

    names_compliant, name_warnings = are_entity_names_dms_compliant(rules, return_report=True)
    if not names_compliant:
        raise exceptions.EntitiesContainNonDMSCompliantCharacters(report=generate_exception_report(name_warnings))

    if methods is None:
        methods = default_class_methods() + default_class_property_methods()

    class_property_pairs = to_class_property_pairs(rules, only_rdfpath=True)

    models: dict[str, ModelMetaclass] = {}
    for class_, properties in class_property_pairs.items():
        # generate fields from define properties
        fields = _properties_to_pydantic_fields(properties)

        if add_extra_fields:
            # store default class to relationship mapping field
            # which is used by the `to_relationship` method
            fields["class_to_asset_mapping"] = (
                dict[str, list[str]],
                Field(
                    define_class_asset_mapping(rules, class_),
                    description="This is a helper field used for generating CDF Asset out of model instance",
                ),
            )

        model = _dictionary_to_pydantic_model(
            model_id=class_,
            model_fields_definition=fields,
            model_name=rules.classes[class_].class_name,
            model_description=rules.classes[class_].description,
            model_methods=methods,
            space=rules.metadata.prefix,
            version=rules.metadata.version,
        )

        models[class_] = model

    return models


def _properties_to_pydantic_fields(
    properties: dict[str, Property],
) -> dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]]:
    """Turns definition of properties into pydantic fields.

    Parameters
    ----------
    properties : dict[str, Property]
        Dictionary of properties

    Returns
    -------
    dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]]
        Dictionary of pydantic fields
    """

    fields: dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]]

    fields = {"external_id": (str, Field(..., alias="external_id"))}

    for property_id, property_ in properties.items():
        field_type = _define_field_type(property_)
        field_definition: dict = {
            "alias": property_.property_name,
            "description": property_.description if property_.description else None,
            # keys below will be available under json_schema_extra
            "property_type": field_type.__name__ if field_type in [EdgeOneToOne, EdgeOneToMany] else "NodeAttribute",
            "property_value_type": property_.expected_value_type.suffix,
            "property_name": property_.property_name,
            "property_id": property_.property_id,
        }

        if field_type.__name__ in [EdgeOneToMany.__name__, list.__name__]:
            field_definition["min_length"] = property_.min_count
            field_definition["max_length"] = property_.max_count

        if not property_.is_mandatory and not property_.default:
            field_definition["default"] = None
        elif property_.default:
            field_definition["default"] = property_.default

        fields[property_id] = (
            field_type,
            Field(**field_definition),  # type: ignore[pydantic-field]
        )
    return fields


def _define_field_type(property_: Property):
    if property_.property_type == "ObjectProperty" and property_.max_count == 1:
        return EdgeOneToOne
    elif property_.property_type == "ObjectProperty":
        return EdgeOneToMany
    elif property_.property_type == "DatatypeProperty" and property_.max_count == 1:
        return cast(ValueTypeMapping, property_.expected_value_type.mapping).python
    else:
        inner_type = cast(ValueTypeMapping, property_.expected_value_type.mapping).python
        return list[inner_type]  # type: ignore[valid-type]


def _dictionary_to_pydantic_model(
    model_id: str,
    model_fields_definition: dict,
    space: str | None = None,
    version: str | None = None,
    model_name: str | None = None,
    model_description: str | None = None,
    model_configuration: ConfigDict | None = None,
    model_methods: list | None = None,
    validators: list | None = None,
) -> type[BaseModel]:
    """Generates pydantic model from dictionary containing definition of fields.
    Additionally, it adds methods to the model and validators.

    Parameters
    ----------
    model_name : str
        Name of the model, typically an id of the class
    model_fields_definition : dict
        Dictionary containing definition of fields
    model_configuration : ConfigDict, optional
        Configuration of pydantic model, by default None
    methods : list, optional
        Methods that work on fields once model is instantiated, by default None
    property_attributes : list, optional
        Property attributed that work on model fields once model is instantiated, by default None
    validators : list, optional
        Any custom validators to be added in addition to base pydantic ones, by default None

    Returns
    -------
    type[BaseModel]
        Pydantic model
    """

    if not model_configuration:
        model_configuration = default_model_configuration(
            external_id=model_id, space=space, version=version, name=model_name, description=model_description
        )

    fields: dict[str, tuple | type[BaseModel]] = {}

    for field_name, value in model_fields_definition.items():
        if isinstance(value, tuple):
            fields[field_name] = value
        # Nested classes
        elif isinstance(value, dict):
            fields[field_name] = (_dictionary_to_pydantic_model(f"{model_id}_{field_name}", value), ...)
        else:
            raise exceptions.FieldValueOfUnknownType(field_name, value)

    model = create_model(model_id, __config__=model_configuration, **fields)  # type: ignore[call-overload]

    if model_methods:
        for method in model_methods:
            try:
                setattr(model, method.__name__, method)
            except AttributeError:
                try:
                    setattr(model, method.fget.__name__, method)
                except AttributeError:
                    setattr(model, method.__func__.fget.__name__, method)

    # any additional validators to be added
    if validators:
        ...

    return model


@classmethod  # type: ignore
@property
def attributes(cls) -> list[str]:
    return [
        field
        for field in cls.model_fields
        if (schema := cls.model_fields[field].json_schema_extra) and schema.get("property_type") == "NodeAttribute"
    ]


@classmethod  # type: ignore
@property
def edges_one_to_one(cls) -> list[str]:
    return [
        field
        for field in cls.model_fields
        if (schema := cls.model_fields[field].json_schema_extra) and schema.get("property_type") == "EdgeOneToOne"
    ]


@classmethod  # type: ignore
@property
def edges_one_to_many(cls) -> list[str]:
    return [
        field
        for field in cls.model_fields
        if (schema := cls.model_fields[field].json_schema_extra) and schema.get("property_type") == "EdgeOneToMany"
    ]


# Define methods that work on model instance
@classmethod  # type: ignore[misc]
def from_dict(cls, dictionary: dict[str, list[str] | str]):
    # wrangle results to dict
    args: dict[str, list[Any] | Any] = {}
    for field in cls.model_fields.values():
        # if field is not required and not in result, skip
        if not field.is_required() and field.alias not in dictionary:
            continue

        # if field is required and not in result, raise error
        if field.is_required() and field.alias not in dictionary:
            raise exceptions.PropertyRequiredButNotProvided(field.alias, cast(str, dictionary["external_id"]))

        # flatten result if field is not edge or list of values
        if field.annotation.__name__ not in [EdgeOneToMany.__name__, list.__name__]:
            if isinstance(dictionary[field.alias], list) and len(dictionary[field.alias]) > 1:
                warnings.warn(
                    exceptions.FieldContainsMoreThanOneValue(
                        field.alias,
                        len(dictionary[field.alias]),
                    ).message,
                    category=exceptions.FieldContainsMoreThanOneValue,
                    stacklevel=2,
                )

            args[field.alias] = dictionary[field.alias][0]
        else:
            args[field.alias] = dictionary[field.alias]

    return cls(**args)


# Define methods that work on model instance
@classmethod  # type: ignore[misc]
def from_graph(
    cls,
    graph: Graph,
    transformation_rules: Rules,
    external_id: URIRef,
    incomplete_instance_allowed: bool = True,
):
    """Method that creates model instance from class instance stored in graph.

    Args:
        graph: Graph containing triples of class instance
        transformation_rules: Transformation rules
        external_id: External id of class instance to be used to instantiate associated pydantic model
        incomplete_instance_allowed: Flag allowing incomplete instances to be queried. Defaults to True.

    Raises:
        exceptions.MissingInstanceTriples: _description_
        exceptions.PropertyRequiredButNotProvided: _description_

    Returns:
        Pydantic model instance
    """

    # build sparql query for given object
    class_ = cls.__name__

    # here properties_optional is set to True in order to also return
    # incomplete class instances so we catch them and raise exceptions
    sparql_construct_query = build_construct_query(
        graph,
        class_,
        transformation_rules,
        class_instances=[external_id],
        properties_optional=incomplete_instance_allowed,
    )
    # In the docs, a construct query is said to return triple
    # Not sure if the triple will be URIRef or Literal
    query_result = cast(Iterable[tuple[URIRef, URIRef, str | URIRef]], graph.query(sparql_construct_query))

    dictionary = triples2dictionary(query_result)[external_id]

    if not dictionary:
        raise exceptions.MissingInstanceTriples(external_id)

    return cls.from_dict(dictionary)


# define methods that creates asset out of model id (default)
def to_asset(
    self,
    data_set_id: int,
    add_system_metadata: bool = True,
    metadata_keys: NeatMetadataKeys | None = None,
    add_labels: bool = True,
) -> Asset:
    """Convert model instance to asset.

    Parameters
    ----------
    add_system_metadata : bool, optional
        Flag indicating to add or not system/neat metadata, by default True
    metadata_keys : NeatMetadataKeys | None, optional
        Definition of system/neat metadata, by default None
    add_labels : bool, optional
        To add or not labels to asset, by default True
    data_set_id : int, optional
        Data set id to which asset belongs to, by default None

    Returns
    -------
    Asset
        Asset instance
    """
    # Needs copy otherwise modifications impact all instances
    if not self.class_to_asset_mapping:
        raise exceptions.ClassToAssetMappingNotDefined(self.__class__.__name__)

    class_instance_dictionary = self.model_dump(by_alias=True)
    adapted_mapping_config = _adapt_mapping_config_by_instance(
        self.external_id, class_instance_dictionary, self.class_to_asset_mapping
    )
    asset = _class_to_asset_instance_dictionary(class_instance_dictionary, adapted_mapping_config)

    # set default metadata keys if not provided
    metadata_keys = NeatMetadataKeys() if metadata_keys is None else metadata_keys

    # add system metadata
    if add_system_metadata:
        _add_system_metadata(self, metadata_keys, asset)

    if add_labels:
        asset["labels"] = [asset["metadata"][metadata_keys.type], "non-historic"]

    return Asset(**asset, data_set_id=data_set_id)


def _add_system_metadata(self, metadata_keys: NeatMetadataKeys, asset: dict):
    asset["metadata"][metadata_keys.type] = self.__class__.__name__
    asset["metadata"][metadata_keys.identifier] = self.external_id
    now = str(datetime.now(UTC))
    asset["metadata"][metadata_keys.start_time] = now
    asset["metadata"][metadata_keys.update_time] = now
    asset["metadata"][metadata_keys.active] = "true"


def _adapt_mapping_config_by_instance(external_id, class_instance_dictionary, mapping_config):
    adapted_mapping_config = {}
    for asset_field, class_properties in mapping_config.items():
        if asset_field != "metadata":
            # We are selecting first property that is available in the graph
            # and add it to corresponding asset field and exit loop
            for property_ in class_properties:
                if class_instance_dictionary.get(property_, None):
                    adapted_mapping_config[asset_field] = property_
                    break

        elif metadata_keys := [
            property_ for property_ in class_properties if class_instance_dictionary.get(property_, None)
        ]:
            adapted_mapping_config["metadata"] = metadata_keys

    # Raise warnings for fields that will miss in asset
    for asset_field in mapping_config:
        if asset_field not in adapted_mapping_config:
            warnings.warn(
                exceptions.FieldNotFoundInInstance(external_id, asset_field).message,
                category=exceptions.FieldNotFoundInInstance,
                stacklevel=2,
            )

    return adapted_mapping_config


def _class_to_asset_instance_dictionary(class_instance_dictionary, mapping_config):
    for key, values in mapping_config.items():
        if key != "metadata":
            mapping_config[key] = class_instance_dictionary.get(values, None)
        else:
            mapping_config[key] = {value: class_instance_dictionary.get(value, None) for value in values}

    return mapping_config


def to_relationship(self, transformation_rules: Rules) -> Relationship:
    """Creates relationship instance from model instance."""
    raise NotImplementedError()


def to_node(
    self, data_model_or_view_id: DMSSchemaComponents | ViewId | None = None, add_class_prefix: bool = False
) -> NodeApply:
    """Creates DMS node from the instance of pydantic model.

    Args:
        data_model_or_view_id: Instance of DataModel or ViewID. Defaults to None.
        add_class_prefix: Whether to add class id (i.e.model name) prefix to external_id of View. Defaults to False.

    Returns:
        Instance of NodeApply containing node information.


    !!! note "Default Behavior"
        If no DataModel or ViewID is passed, then the default behavior is to create node
        using View information which is by default stored under `model_json_schema` attribute of pydantic model.
    !!! note "Limitations"
        Currently adding class prefix is only possible if the node is created if ViewID is passed or
        if pydantic model already contains View information (which default behavior).
    """

    if isinstance(data_model_or_view_id, DMSSchemaComponents):
        return _to_node_using_data_model(self, data_model_or_view_id, add_class_prefix)
    elif isinstance(data_model_or_view_id, ViewId):
        if not data_model_or_view_id.space:
            raise exceptions.SpaceNotDefined()
        if not data_model_or_view_id.external_id:
            raise exceptions.ViewExternalIdNotDefined()
        if not data_model_or_view_id.version:
            raise exceptions.ViewVersionNotDefined()
        return _to_node_using_view_id(self, data_model_or_view_id)
    else:
        space = self.model_json_schema().get("space", None)
        external_id = self.model_json_schema().get("external_id", None)
        version = self.model_json_schema().get("version", None)

        if not space:
            raise exceptions.SpaceNotDefined()
        if not external_id:
            raise exceptions.ViewExternalIdNotDefined()
        if not version:
            raise exceptions.ViewVersionNotDefined()

        return _to_node_using_view_id(self, ViewId(space, external_id, version))


def _to_node_using_view_id(self, view_id: ViewId) -> NodeApply:
    attributes: dict = {
        attribute: (
            getattr(self, attribute).isoformat()
            if isinstance(getattr(self, attribute), date)
            else getattr(self, attribute)
        )
        for attribute in self.attributes
    }

    edges_one_to_one: dict = {}

    for edge in self.edges_one_to_one:
        if external_id := getattr(self, edge):
            edges_one_to_one[edge] = {
                "space": self.model_json_schema()["space"],
                "externalId": external_id,
            }

    return NodeApply(
        space=self.model_json_schema()["space"],
        external_id=self.external_id,
        sources=[
            NodeOrEdgeData(
                source=view_id,
                properties=attributes | edges_one_to_one,
            )
        ],
    )


def _to_node_using_data_model(self, data_model, add_class_prefix) -> NodeApply:
    view_id: str = f"{self.model_json_schema()['space']}:{type(self).__name__}"

    if not set(self.attributes + self.edges_one_to_one).issubset(set(data_model.views[view_id].properties.keys())):
        raise exceptions.InstancePropertiesNotMatchingViewProperties(
            self.__class__.__name__,
            self.attributes + self.edges_one_to_one + self.edges_one_to_many,
            list(data_model.views[view_id].properties.keys()),
        )

    attributes: dict = {
        attribute: (
            getattr(self, attribute).isoformat()
            if isinstance(getattr(self, attribute), date)
            else getattr(self, attribute)
        )
        for attribute in self.attributes
    }
    if add_class_prefix:
        edges_one_to_one: dict = {}
        dm_view = data_model.views[view_id]
        if dm_view.properties:
            for edge in self.edges_one_to_one:
                mapped_property = dm_view.properties[edge]
                if isinstance(mapped_property, MappedPropertyApply):
                    object_view = mapped_property.source
                if object_view:
                    object_class_name = object_view.external_id

                    if external_id := getattr(self, edge):
                        edges_one_to_one[edge] = {
                            "space": data_model.views[view_id].space,
                            "externalId": add_class_prefix_to_xid(
                                class_name=object_class_name,
                                external_id=external_id,
                            ),
                        }
    else:
        edges_one_to_one = {}

        for edge in self.edges_one_to_one:
            if external_id := getattr(self, edge):
                edges_one_to_one[edge] = {
                    "space": self.model_json_schema()["space"],
                    "externalId": external_id,
                }

    return NodeApply(
        space=data_model.views[view_id].space,
        external_id=self.external_id,
        sources=[
            NodeOrEdgeData(
                source=data_model.views[view_id].as_id(),
                properties=attributes | edges_one_to_one,
            )
        ],
    )


def to_edge(self, data_model: DMSSchemaComponents, add_class_prefix: bool = False) -> list[EdgeApply]:
    """Creates DMS edge from pydantic model."""
    edges: list[EdgeApply] = []

    def is_external_id_valid(external_id: str) -> bool:
        # should match "^[^\x00]{1,255}$" and not be None or none
        if external_id == "None" or external_id == "none":
            return False
        return bool(re.match(r"^[^\x00]{1,255}$", external_id))

    class_name: str = type(self).__name__
    view_id: str = f"{self.model_json_schema()['space']}:{class_name}"

    for edge_one_to_many in self.edges_one_to_many:
        edge_type_id = f"{class_name}.{edge_one_to_many}"
        if end_node_ids := getattr(self, edge_one_to_many):
            for end_node_id in end_node_ids:
                if not is_external_id_valid(end_node_id):
                    warnings.warn(
                        message=exceptions.EdgeConditionUnmet(edge_one_to_many).message,
                        category=exceptions.EdgeConditionUnmet,
                        stacklevel=2,
                    )
                end_node_external_id = end_node_id
                if add_class_prefix:
                    end_node_class_name = _get_end_node_class_name(data_model.views[view_id], edge_one_to_many)
                    if end_node_class_name:
                        end_node_external_id = add_class_prefix_to_xid(end_node_class_name, end_node_id)
                    else:
                        warnings.warn(
                            message=exceptions.EdgeConditionUnmet(edge_one_to_many).message,
                            category=exceptions.EdgeConditionUnmet,
                            stacklevel=2,
                        )

                external_id = f"{self.external_id}.{edge_one_to_many}.{end_node_external_id}"
                edge = EdgeApply(
                    space=data_model.views[view_id].space,
                    external_id=external_id if len(external_id) < 256 else create_sha256_hash(external_id),
                    type=(data_model.views[view_id].space, edge_type_id),
                    start_node=(data_model.views[view_id].space, self.external_id),
                    end_node=(data_model.views[view_id].space, end_node_external_id),
                )
                edges.append(edge)
    return edges


def _get_end_node_class_name(view: ViewApply, edge: str) -> str | None:
    """Get the class name of the end node of an edge."""
    if view.properties is None:
        return None
    mapped_instance = view.properties[edge]
    if isinstance(mapped_instance, SingleHopConnectionDefinitionApply) and mapped_instance.source:
        return mapped_instance.source.external_id
    return None


@classmethod  # type: ignore
@property
def model_name(cls) -> str | None:
    """Returns the name of the model if one exists"""
    return cls.model_json_schema().get("class_name", None)


@classmethod  # type: ignore
@property
def model_external_id(cls) -> str | None:
    """Returns the external id of the model if one exists"""
    return cls.model_json_schema().get("class_id", None)


@classmethod  # type: ignore
@property
def model_description(cls) -> str | None:
    """Returns the description of the model if one exists"""
    return cls.model_json_schema().get("description", None)


@classmethod  # type: ignore
def get_field_description(cls, field_id: str) -> str | None:
    """Returns description of the field if one exists"""
    if field_id in cls.model_fields:
        return cls.model_fields[field_id].description
    else:
        return None


@classmethod  # type: ignore
def get_field_name(cls, field_id: str) -> str | None:
    """Returns name of the field if one exists"""
    if field_id in cls.model_fields and cls.model_fields[field_id].json_schema_extra:
        if "property_name" in cls.model_fields[field_id].json_schema_extra:
            return cls.model_fields[field_id].json_schema_extra["property_name"]
    return None


def add_class_prefix_to_xid(class_name: str, external_id: str) -> str:
    """Adds class name as prefix to the external_id"""
    return f"{class_name}_{external_id}"



""" """

import re
import sys
from collections import Counter
from functools import total_ordering
from typing import ClassVar, Literal

from pydantic import BaseModel, field_validator, model_serializer

from cognite.neat.rules import exceptions

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self


class TransformationRuleType(StrEnum):
    rdfpath = "rdfpath"
    rawlookup = "rawlookup"
    sparql = "sparql"


class Lookup(StrEnum):
    table = "table"
    key = "key"
    value = "value"  # type: ignore


class EntityTypes(StrEnum):
    class_ = "class"
    property_ = "property"
    undefined = "undefined"


# FOR PARSING STRINGS:
PREFIX_REGEX = r"[a-zA-Z]+[a-zA-Z0-9-_.]*[a-zA-Z0-9]+"
SUFFIX_REGEX = r"[a-zA-Z0-9-_.]+[a-zA-Z0-9]|[-_.]*[a-zA-Z0-9]+"
VERSION_REGEX = r"[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?"

ENTITY_ID_REGEX = rf"{PREFIX_REGEX}:({SUFFIX_REGEX})"
ENTITY_ID_REGEX_COMPILED = re.compile(rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})$")
VERSIONED_ENTITY_REGEX_COMPILED = re.compile(
    rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})\(version=(?P<version>{VERSION_REGEX})\)$"
)
CLASS_ID_REGEX = rf"(?P<{EntityTypes.class_}>{ENTITY_ID_REGEX})"
CLASS_ID_REGEX_COMPILED = re.compile(rf"^{CLASS_ID_REGEX}$")
PROPERTY_ID_REGEX = rf"\((?P<{EntityTypes.property_}>{ENTITY_ID_REGEX})\)"

# traversal direction
DIRECTION_REGEX = r"(?P<direction>(->|<-))"

# steps
STEP_REGEX = rf"((->|<-){CLASS_ID_REGEX}({PROPERTY_ID_REGEX})?)"
STEP_REGEX_COMPILED = re.compile(STEP_REGEX)
STEP_CLASS_REGEX_COMPILED = re.compile(rf"(^{DIRECTION_REGEX}{CLASS_ID_REGEX})$")
STEP_CLASS_AND_PROPERTY_REGEX_COMPILED = re.compile(rf"(^{DIRECTION_REGEX}{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}$)")


_traversal = "traversal"
ORIGIN_REGEX = rf"(?P<origin>{ENTITY_ID_REGEX})"

HOP_REGEX_COMPILED = re.compile(rf"^{ORIGIN_REGEX}(?P<{_traversal}>{STEP_REGEX}+)$")

# grabbing specific property for a class, property can be either object, annotation or data property
SINGLE_PROPERTY_REGEX_COMPILED = re.compile(rf"^{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}$")

# grabbing all properties for a class
ALL_PROPERTIES_REGEX_COMPILED = re.compile(rf"^{CLASS_ID_REGEX}\(\*\)$")

ALL_TRAVERSAL_REGEX_COMPILED = (
    rf"({CLASS_ID_REGEX}\(\*\)|{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}|{ORIGIN_REGEX}(?P<{_traversal}>{STEP_REGEX}+))"
)

TABLE_REGEX_COMPILED = re.compile(
    rf"^(?P<{Lookup.table}>{SUFFIX_REGEX})\((?P<{Lookup.key}>{SUFFIX_REGEX}),\s*(?P<{Lookup.value}>{SUFFIX_REGEX})\)$"
)


StepDirection = Literal["source", "target", "origin"]
_direction_by_symbol: dict[str, StepDirection] = {"->": "target", "<-": "source"}
_symbol_by_direction: dict[StepDirection, str] = {"source": "<-", "target": "->"}

Undefined = type(object())
Unknown = type(object())


# mypy does not like the sentinel value, and it is not possible to ignore only the line with it below.
# so we ignore all errors beyond this point.
# mypy: ignore-errors
@total_ordering
class Entity(BaseModel, arbitrary_types_allowed=True):
    """Entity is a class or property in OWL/RDF sense."""

    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: str | Undefined = Undefined
    suffix: str | Unknown
    version: str | None = None
    name: str | None = None
    description: str | None = None

    def __lt__(self, other: object) -> bool:
        if type(self) is not type(other) or not isinstance(other, Entity):
            return NotImplemented
        return self.versioned_id < other.versioned_id

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other) or not isinstance(other, Entity):
            return NotImplemented
        return self.versioned_id == other.versioned_id

    def __hash__(self) -> int:
        return hash(self.versioned_id)

    def as_non_versioned_entity(self) -> Self:
        return self.from_string(f"{self.prefix}:{self.suffix}")

    @property
    def id(self) -> str:
        if self.suffix is Unknown:
            return "#N/A"
        elif self.prefix is Undefined:
            return self.suffix
        else:
            return f"{self.prefix}:{self.suffix}"

    @property
    def versioned_id(self) -> str:
        if self.version is None:
            return self.id
        else:
            return f"{self.id}(version={self.version})"

    @property
    def space(self) -> str:
        """Returns entity space in CDF."""
        return self.prefix

    @property
    def external_id(self) -> str:
        """Returns entity external id in CDF."""
        return self.suffix

    def __repr__(self):
        return self.versioned_id

    def __str__(self):
        return self.versioned_id

    @classmethod
    def from_string(cls, entity_string: str, base_prefix: str | None = None) -> Self:
        if entity_string == "#N/A":
            return cls(prefix=Undefined, suffix=Unknown)
        elif result := VERSIONED_ENTITY_REGEX_COMPILED.match(entity_string):
            return cls(
                prefix=result.group("prefix"),
                suffix=result.group("suffix"),
                version=result.group("version"),
            )
        elif result := ENTITY_ID_REGEX_COMPILED.match(entity_string):
            return cls(prefix=result.group("prefix"), suffix=result.group("suffix"))
        elif base_prefix and re.match(SUFFIX_REGEX, entity_string) and re.match(PREFIX_REGEX, base_prefix):
            return cls(prefix=base_prefix, suffix=entity_string)
        else:
            raise ValueError(f"{cls.__name__} is expected to be prefix:suffix, got {entity_string}")

    @classmethod
    def from_list(cls, entity_strings: list[str], base_prefix: str | None = None) -> list[Self]:
        return [
            cls.from_string(entity_string=entity_string, base_prefix=base_prefix) for entity_string in entity_strings
        ]


class Step(BaseModel):
    class_: Entity
    property: Entity | None = None  # only terminal step has property
    direction: StepDirection

    @classmethod
    def from_string(cls, raw: str, **kwargs) -> Self:
        if result := STEP_CLASS_AND_PROPERTY_REGEX_COMPILED.match(raw):
            return cls(
                class_=Entity.from_string(result.group(EntityTypes.class_)),
                property=Entity.from_string(result.group(EntityTypes.property_)),
                direction=_direction_by_symbol[result.group("direction")],
                **kwargs,
            )
        elif result := STEP_CLASS_REGEX_COMPILED.match(raw):
            return cls(
                class_=Entity.from_string(result.group(EntityTypes.class_)),
                direction=_direction_by_symbol[result.group("direction")],
            )  # type: ignore
        msg = f"Invalid step {raw}, expected in one of the following forms:"
        msg += " ->prefix:suffix, <-prefix:suffix, ->prefix:suffix(prefix:suffix) or <-prefix:suffix(prefix:suffix)"
        raise ValueError(msg)

    def __str__(self) -> str:
        if self.property:
            return f"{self.class_}({self.property})"
        else:
            return f"{_symbol_by_direction[self.direction]}{self.class_}"

    def __repr__(self) -> str:
        return self.__str__()


class Traversal(BaseModel):
    class_: Entity

    def __str__(self) -> str:
        return f"{self.class_}"

    def __repr__(self) -> str:
        return self.__str__()

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)


class SingleProperty(Traversal):
    property: Entity

    @classmethod
    def from_string(cls, class_: str, property_: str) -> Self:
        return cls(class_=Entity.from_string(class_), property=Entity.from_string(property_))

    def __str__(self) -> str:
        return f"{self.class_}({self.property})"


class AllReferences(Traversal):
    @classmethod
    def from_string(cls, class_: str) -> Self:
        return cls(class_=Entity.from_string(class_))


class AllProperties(Traversal):
    @classmethod
    def from_string(cls, class_: str) -> Self:
        return cls(class_=Entity.from_string(class_))

    def __str__(self) -> str:
        return f"{self.class_}(*)"


class Origin(BaseModel):
    class_: Entity

    @field_validator("class_", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class Hop(Traversal):
    """Multi or single hop traversal through graph"""

    traversal: list[Step]

    @classmethod
    def from_string(cls, class_: str, traversal: str | list[Step]) -> Self:
        return cls(
            class_=Entity.from_string(class_),
            traversal=(
                [Step.from_string(result[0]) for result in STEP_REGEX_COMPILED.findall(traversal)]
                if isinstance(traversal, str)
                else traversal
            ),
        )

    def __str__(self) -> str:
        return f"{self.class_}{''.join([str(step) for step in self.traversal])}"


class TableLookup(BaseModel):
    name: str
    key: str
    value: str


class Rule(BaseModel):
    pass


class Query(BaseModel):
    query: str


class RDFPath(Rule):
    traversal: SingleProperty | AllProperties | AllReferences | Hop

    def __str__(self) -> str:
        return f"{self.traversal}"

    def __repr__(self) -> str:
        return self.__str__()

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)


class RawLookup(RDFPath):
    table: TableLookup


class SPARQLQuery(RDFPath):
    traversal: Query


def parse_traversal(raw: str) -> AllReferences | AllProperties | SingleProperty | Hop:
    if result := CLASS_ID_REGEX_COMPILED.match(raw):
        return AllReferences.from_string(class_=result.group(EntityTypes.class_))
    elif result := ALL_PROPERTIES_REGEX_COMPILED.match(raw):
        return AllProperties.from_string(class_=result.group(EntityTypes.class_))
    elif result := SINGLE_PROPERTY_REGEX_COMPILED.match(raw):
        return SingleProperty.from_string(
            class_=result.group(EntityTypes.class_), property_=result.group(EntityTypes.property_)
        )
    elif result := HOP_REGEX_COMPILED.match(raw):
        return Hop.from_string(class_=result.group("origin"), traversal=result.group(_traversal))
    else:
        raise exceptions.NotValidRDFPath(raw).to_pydantic_custom_error()


def parse_table_lookup(raw: str) -> TableLookup:
    if result := TABLE_REGEX_COMPILED.match(raw):
        return TableLookup(
            name=result.group(Lookup.table), key=result.group(Lookup.key), value=result.group(Lookup.value)
        )
    raise exceptions.NotValidTableLookUp(raw).to_pydantic_custom_error()


def parse_rule(rule_raw: str, rule_type: TransformationRuleType | None) -> RDFPath:
    match rule_type:
        case TransformationRuleType.rdfpath:
            rule_raw = rule_raw.replace(" ", "")
            return RDFPath(traversal=parse_traversal(rule_raw))
        case TransformationRuleType.rawlookup:
            rule_raw = rule_raw.replace(" ", "")
            if Counter(rule_raw).get("|") != 1:
                raise exceptions.NotValidRAWLookUp(rule_raw).to_pydantic_custom_error()
            traversal, table_lookup = rule_raw.split("|")
            return RawLookup(traversal=parse_traversal(traversal), table=parse_table_lookup(table_lookup))
        case TransformationRuleType.sparql:
            return SPARQLQuery(traversal=Query(query=rule_raw))
        case None:
            raise ValueError("Rule type must be specified")


def is_valid_rule(rule_type: TransformationRuleType, rule_raw: str) -> bool:
    is_valid_rule = {TransformationRuleType.rdfpath: is_rdfpath, TransformationRuleType.rawlookup: is_rawlookup}[
        rule_type
    ]
    return is_valid_rule(rule_raw)


def is_rdfpath(raw: str) -> bool:
    try:
        parse_traversal(raw)
    except ValueError:
        return False
    return True


def is_rawlookup(raw: str) -> bool:
    try:
        parse_rule(raw, TransformationRuleType.rawlookup)
    except ValueError:
        return False
    return True



"""This module contains the definition of errors and warnings raised when dealing
with TransformationRules object. This includes underlying pydantic model, actual transformation rules
handling (such `rdfpath`), and rules loaders, parsers and exporters.

\nThe errors and warning are grouped by means of error codes:\n
- 0 - 99: errors and warnings raised when dealing with TransformationRules pydantic model
- 100 - 199: errors and warnings raised when parsing actual transformation rules, i.e. `rdfpath`
- 200 - 299: errors and warnings raised when dealing TransformationRules importers
- 300 - 399: errors and warnings raised when dealing TransformationRules parsers
- 400 - 499: errors and warnings raised when dealing TransformationRules exporters

"""

from typing import Any

from cognite.client.data_classes.data_modeling import ContainerId, DataModelId, ViewId
from rdflib import Namespace, URIRef

from cognite.neat.constants import DEFAULT_DOCS_URL
from cognite.neat.exceptions import NeatException, NeatWarning

DOCS_BASE_URL = f"{DEFAULT_DOCS_URL}api/exceptions.html#{__name__}"


class MultipleExceptions(NeatException):
    """This is used to gather multiple errors."""

    def __init__(self, errors: list[NeatException], verbose: bool = False):
        self.errors = errors
        self.message = f"Multiple errors occurred: {self.errors}"
        if verbose:
            self.message += f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        super().__init__(self.message)


################################################################################################
# RULES MODEL REPRESENTATION: 100 - 199 ########################################################
################################################################################################

# Exceptions:


class PrefixRegexViolation(NeatException):
    """Prefix, which is in the 'Metadata' sheet, does not respect defined regex expression

    Args:
        prefix: prefix that raised exception
        regex_expression: regex expression against which prefix is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if prefix in the 'Metadata' sheet contains any illegal characters and
        respects the regex expression.

    """

    type_: str = "PrefixRegexViolation"
    code: int = 0
    description: str = "Prefix, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If prefix is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if prefix in the 'Metadata' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, prefix: str, regex_expression: str, verbose: bool = False):
        self.prefix = prefix
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid prefix '{self.prefix}' stored in 'Metadata' sheet, it must obey regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PrefixMissing(NeatException):
    """Prefix, which is in the 'Metadata' sheet, is missing.

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "PrefixMissing"
    code: int = 0
    description: str = "Prefix is missing from the 'Metadata' sheet."
    example: str = "There is no prefix in the 'Metadata' sheet."
    fix: str = "Specify the prefix if prefix in the 'Metadata' sheet."

    def __init__(self, verbose: bool = False):
        self.message = (
            f"Missing prefix stored in 'Metadata' sheet."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class CDFSpaceRegexViolation(NeatException):
    """cdfSpaceName, which is in the 'Metadata' sheet, does not respect defined regex expression

    Args:
        cdf_space_name: cdf_space_name that raised exception
        regex_expression: regex expression against which cdf_space_name is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if cdfSpaceName in the 'Metadata' sheet contains any illegal characters and
        respects the regex expression.

    """

    type_: str = "CDFSpaceRegexViolation"
    code: int = 1
    description: str = "cdfSpaceName, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If cdfSpaceName is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if cdfSpaceName in the 'Metadata' sheet "
        "contains any illegal characters and respects the regex expression"
    )

    def __init__(self, cdf_space_name: str, regex_expression: str, verbose: bool = False):
        self.cdf_space_name = cdf_space_name
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid cdfSpaceName '{self.cdf_space_name}' stored in 'Metadata' sheet, "
            f"it must obey regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MetadataSheetNamespaceNotValidURL(NeatException):
    """namespace, which is in the 'Metadata' sheet, does not respect defined regex expression

    Args:
        namespace: namespace that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if `namespace` in the `Metadata` sheet is properly constructed as valid URL
        containing only allowed characters.

    """

    type_: str = "MetadataSheetNamespaceNotValidURL"
    code: int = 2
    description: str = "namespace, which is in the 'Metadata' sheet, is not valid URL"
    example: str = "If we have 'authority:namespace' as namespace as it is not a valid URL this error will be raised"
    fix: str = (
        "Check if 'namespace' in the 'Metadata' sheet is properly "
        "constructed as valid URL containing only allowed characters"
    )

    def __init__(self, namespace: str, verbose: bool = False):
        self.namespace = namespace

        self.message = (
            f"Invalid namespace '{self.namespace}' stored in 'Metadata' sheet, it must be valid URL!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MetadataSheetNamespaceNotDefined(NeatException):
    """namespace, which is in the 'Metadata' sheet, is not defined

    Args:
        namespace: namespace that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if `namespace` in the `Metadata` sheet is properly constructed as valid URL
        containing only allowed characters.

    """

    type_ = "MetadataSheetNamespaceNotDefined"
    code: int = 2
    description: str = "namespace, which is in the 'Metadata' sheet, is missing"
    example: str = "Example of a valid namespace 'http://www.w3.org/ns/sparql#'"
    fix: str = "Define the 'namespace' in the 'Metadata' sheet."

    def __init__(self, verbose: bool = False):
        self.message = (
            f"Missing namespace  in 'Metadata' sheet." f"\nFor more information visit: {DOCS_BASE_URL}.{self.type_}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class DataModelIdRegexViolation(NeatException):
    """dataModelName, which is in the 'Metadata' sheet, does not respect defined regex expression

    Args:
        data_model_name: data_model_name that raised exception
        regex_expression: regex expression against which data_model_name is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if `dataModelName` in the `Metadata` sheet contains any illegal
        characters and respects the regex expression

    """

    type_: str = "DataModelIdRegexViolation"
    code: int = 3
    description: str = "external_id, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If external_id is set to 'power grid data model', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if external_id in the 'Metadata' sheet contains any illegal "
        "characters and respects the regex expression"
    )

    def __init__(self, data_model_id: str, regex_expression: str, verbose: bool = False):
        self.data_model_id = data_model_id
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid data model external_id '{self.data_model_id}' stored in 'Metadata' sheet, "
            f"it must obey regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class VersionRegexViolation(NeatException):
    """version, which is in the 'Metadata' sheet, does not respect defined regex expression

    Args:
        version: version that raised exception
        regex_expression: regex expression against which version is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check if `version` in the `Metadata` sheet contains any illegal
        characters and respects the regex expression

    """

    type_: str = "VersionRegexViolation"
    code: int = 4
    description: str = "version, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If version is set to '1.2.3 alpha4443', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if version in the 'Metadata' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, version: str, regex_expression: str, verbose: bool = False):
        self.version = version
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid version '{self.version}' stored in 'Metadata' sheet, it must obey regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ClassSheetClassIDRegexViolation(NeatException):
    """Class ID, which is stored in the column 'Class' in the 'Classes' sheet, does not
    respect defined regex expression

    Args:
        class_id: class_id that raised exception
        regex_expression: regex expression against which class_id is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check definition of class ids in 'Class' column in 'Classes' sheet and
        make sure to respect the regex expression by removing any illegal characters

    """

    type_: str = "ClassSheetClassIDRegexViolation"
    code: int = 5
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Classes' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If class id is set to 'Class 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of class ids in 'Class' column in 'Classes' sheet and "
        "make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, class_id: str, regex_expression: str, verbose: bool = False):
        self.class_id = class_id
        self.regex_expression = regex_expression

        self.message = (
            f"Class id '{self.class_id}' stored in 'Class' column in 'Classes' "
            f"sheet violates regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ClassIDMissing(NeatException):
    """Class ID, which is stored in the column 'Class' in the 'Classes' sheet, is either
    missing or did not satisfied regex expression

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure that class id is provided and respects regex expression
    """

    type_: str = "ClassIDMissing"
    code: int = 6
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Classes' sheet,"
        " is either missing or did not satisfied regex expression"
    )
    example: str = ""
    fix: str = "Make sure that class id is provided and respects regex expression"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Class id is missing, it failed validation either because it has"
            " not been provided or because it did not respect regex expression!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertiesSheetClassIDRegexViolation(NeatException):
    """Class ID, which is stored in the column 'Class' in the 'Properties' sheet, does
    not respect defined regex expression

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
        class_id: class id that raised exception
        regex_expression: regex expression against which class id is checked

    Notes:
        Check definition of class ids in `Class` column in `Properties` sheet and make
        sure to respect the regex expression by removing any illegal characters
    """

    type_: str = "PropertiesSheetClassIDRegexViolation"
    code: int = 7
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Properties' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If class id is set to 'Class 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of class ids in 'Class' column in 'Properties' sheet "
        "and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, class_id: str, regex_expression: str, verbose: bool = False):
        self.class_id = class_id
        self.regex_expression = regex_expression

        self.message = (
            f"Class id '{self.class_id}' stored in 'Class' column in 'Properties' "
            f"sheet violates regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertyIDRegexViolation(NeatException):
    """Property ID, which is stored in the column 'Property' in the 'Properties' sheet, does
    not respect defined regex expression

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
        property_id: property id that raised exception
        regex_expression: regex expression against which property id is checked

    Notes:
        Check definition of class ids in `Property` column in `Properties` sheet and make
        sure to respect the regex expression by removing any illegal characters
    """

    type_: str = "PropertyIDRegexViolation"
    code: int = 8
    description: str = (
        "Property ID, which is stored in the column 'Property' "
        "in the 'Properties' sheet, does not respect defined regex expression"
    )
    example: str = (
        "If property id is set to 'property 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of property ids in 'Property' column in 'Properties' sheet"
        " and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, property_id: str, regex_expression: str, verbose: bool = False):
        self.property_id = property_id
        self.regex_expression = regex_expression

        self.message = (
            f"Property id '{self.property_id}' stored in 'Property' "
            f"column in 'Properties' sheet violates regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ValueTypeIDRegexViolation(NeatException):
    """Value type, which is stored in the column 'Type' in the 'Properties' sheet, does
    not respect defined regex expression

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
        value_type: value type that raised exception
        regex_expression: regex expression against which value type is checked

    Notes:
        Check definition of class ids in `Type` column in `Properties` sheet and make
        sure to respect the regex expression by removing any illegal characters
    """

    type_: str = "ValueTypeIDRegexViolation"
    code: int = 9
    description: str = (
        "Value type, which is stored in the column 'Type' in the 'Properties' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If value type is set to 'date time', while regex expression does not"
        " allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of value types in 'Type' column in 'Properties' sheet"
        " and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, value_type: str, regex_expression: str, verbose: bool = False):
        self.value_type = value_type
        self.regex_expression = regex_expression

        self.message = (
            f"Value type '{self.value_type}' stored in 'Type' column in "
            f"'Properties' sheet violates regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MissingTypeValue(NeatException):
    """Value type, which is stored in the column 'Type' in the 'Properties' sheet, is missing

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure to define value type in `Type` column in `Properties` sheet
    """

    type_: str = "MissingTypeValue"
    code: int = 10
    description: str = "Value type, which is stored in the column 'Type' in the 'Properties' sheet, is missing"
    example: str = "If value type is not set, this error will be raised"
    fix: str = "Make sure to define value type in 'Type' column in 'Properties' sheet"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Value type, which is stored in the column 'Type' in the 'Properties' sheet, is missing"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertyIDMissing(NeatException):
    """Property ID, which is stored in the column 'Property' in the 'Properties' sheet,
    is either missing or did not satisfied regex expression

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure to define value type in `Type` column in `Properties` sheet
    """

    type_: str = "PropertyIDMissing"
    code: int = 11
    description: str = (
        "Property ID, which is stored in the column 'Property' in the 'Properties' sheet,"
        " is either missing or did not satisfied regex expression"
    )
    example: str = ""
    fix: str = "Make sure that property id is provided and respects regex expression"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Property id is missing, validator for property id failed either"
            " due to lack of property id or due to not respecting regex expression!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class RuleTypeProvidedButRuleMissing(NeatException):
    """This error occurs when transformation rule type is provided but actual
    transformation rule is missing

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
        property_id: property id which is missing transformation rule
        class_id: class id for which property id is defined
        rule_type: rule type that is provided for property id

    Notes:
        If you provide rule type you must provide rule as well! Otherwise remove rule
        type if no transformation rule is needed
    """

    type_: str = "RuleTypeProvidedButRuleMissing"
    code: int = 12
    description: str = (
        "This error occurs when transformation rule type is provided but actual transformation rule is missing"
    )
    example: str = ""
    fix: str = (
        "If you provide rule type you must provide rule as well! "
        "Otherwise remove rule type if no transformation rule is needed"
    )

    def __init__(self, property_id: str, class_id: str, rule_type: str, verbose: bool = False):
        self.message = (
            f"Rule type '{rule_type}' provided for property '{property_id}' "
            f"in class '{class_id}' but rule is not provided!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertyDefinedForUndefinedClass(NeatException):
    """Property defined for a class that has not been defined in the 'Classes' sheet

    Args:
        property_id: property id that is defined for undefined class
        class_id: class id that is undefined
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure to define all classes in the 'Classes' sheet before defining properties
        for them in the `Properties` sheet
    """

    type_: str = "PropertyDefinedForUndefinedClass"
    code: int = 13
    description: str = "Property defined for a class that has not been defined in the 'Classes' sheet"
    example: str = (
        "If property 'someProperty' is defined for class 'Class 1' in the 'Properties' sheet, "
        "while 'Class 1' has not been defined in the 'Classes' sheet,"
        " this error will be raised"
    )
    fix: str = (
        "Make sure to define all classes in the 'Classes' sheet before defining properties for them"
        " in the 'Properties' sheet"
    )

    def __init__(self, property_id: str, class_id: str, verbose: bool = False):
        self.property_id = property_id
        self.class_id = class_id

        self.message = (
            f"Class <{self.class_id}> to which property {self.property_id}> is being defined"
            " is not define in the 'Classes' sheet!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MetadataSheetMissingOrFailedValidation(NeatException):
    """Metadata sheet is missing or it failed validation for one or more fields

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure to define compliant Metadata sheet before proceeding
    """

    type_: str = "MetadataSheetMissingOrFailedValidation"
    code: int = 14
    description: str = "Metadata sheet is missing or it failed validation for one or more fields"
    example: str = ""
    fix: str = "Make sure to define compliant Metadata sheet before proceeding"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Metadata sheet is missing or it failed validation for one or more fields!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class FiledInMetadataSheetMissingOrFailedValidation(NeatException):
    """One of the mandatory fields in Metadata sheet is missing or it failed validation

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure to define compliant field in Metadata sheet before proceeding
    """

    type_: str = "FiledInMetadataSheetMissingOrFailedValidation"
    code: int = 15
    description: str = "One of the mandatory fields in Metadata sheet is missing or it failed validation"
    example: str = ""
    fix: str = "Make sure to define compliant field in Metadata sheet before proceeding"

    def __init__(self, missing_field: str, verbose: bool = False):
        self.message = (
            f"Field {missing_field} is missing in the 'Metadata' sheet or it failed validation!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PrefixesRegexViolation(NeatException):
    """Prefix(es), which are in the 'Prefixes' sheet, do(es) not respect defined regex expression

    Args:
        prefixes: list of prefixes that violate regex expression
        regex_expression: regex expression that is violated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Check if prefixes in the `Prefixes` sheet contains any illegal characters and respects the regex expression
    """

    type_: str = "PrefixesRegexViolation"
    code: int = 16
    description: str = "Prefix(es), which are in the 'Prefixes' sheet, do(es) not respect defined regex expression"
    example: str = (
        "If prefix is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if prefixes in the 'Prefixes' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, prefixes: list[str], regex_expression: str, verbose: bool = False):
        self.prefixes = prefixes
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid prefix(es) {', '.join(self.prefixes)} stored in the 'Prefixes' sheet, "
            f"it/they must obey regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PrefixesSheetNamespaceNotValidURL(NeatException):
    """Namespace(es), which are/is in the 'Prefixes' sheet, are/is not valid URL(s)

    Args:
        namespaces: list of namespaces that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Check if `namespaces` in the `Prefixes` sheet are properly constructed as valid
       URLs containing only allowed characters
    """

    type_: str = "PrefixesSheetNamespaceNotValidURL"
    code: int = 17
    description: str = "Namespace(es), which are/is in the 'Prefixes' sheet, are/is not valid URLs"
    example: str = "If we have 'authority:namespace' as namespace as it is not a valid URL this error will be raised"
    fix: str = (
        "Check if 'namespaces' in the 'Prefixes' sheet are properly "
        "constructed as valid URLs containing only allowed characters"
    )

    def __init__(self, namespaces: list[str], verbose: bool = False):
        self.namespaces = namespaces

        self.message = (
            f"Invalid namespace(es) {', '.join(self.namespaces)} stored in the 'Prefixes' sheet, "
            f"it/they must be valid URLs!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ValueTypeNotDefinedAsClass(NeatException):
    """Expected value type, which is stored in the column 'Type' in the 'Properties' sheet,
    is not defined in the 'Classes' sheet. This error occurs when property is defined as
    an edge between two classes, of which one is not defined

    Args:
        expected_value_type: expected value type that raised exception
        property_id: property id that has expected value type that raised exception
        class_id: class id for which property is defined
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure to define all of the classes in the `Classes` sheet before defining
        properties that expect them as value types
    """

    type_: str = "ValueTypeNotDefinedAsClass"
    code: int = 18
    description: str = (
        "Expected value type, which is stored in the column 'Type' in the 'Properties'"
        " sheet, is not defined in the 'Classes' sheet. "
        "This error occurs when property is defined as an edge between two classes, of which one is not defined"
    )
    example: str = (
        "We have 'Class1' which has property 'edgeClass1Class2' linking it to 'Class2', thus"
        "expected value of 'edgeClass1Class2' is 'Class2'. However, 'Classes' sheet only contains"
        " 'Class1', while 'Class2' is not defined. Under this given circumstance, this error will be raised!"
    )

    fix: str = (
        "Make sure to define all of the classes in the 'Classes' sheet before defining "
        "properties that expect them as value types"
    )

    def __init__(self, class_id: str, property_id: str, expected_value_type: str, verbose: bool = False):
        self.message = (
            f"Property {property_id} defined for class {class_id} has"
            f" value type {expected_value_type} which is not defined as a class in the 'Classes' sheet!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class UndefinedObjectsAsExpectedValueTypes(NeatException):
    """Expected value types, which are stored in the column 'Type' in the 'Properties'
    sheet, are classes that exist in the 'Classes' sheet but for which no properties are defined
    in the 'Properties' sheet.

    Args:
        undefined_objects: list of undefined objects that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure to define properties for classes from 'Classes' sheet before defining
        properties that expect them as value types
    """

    type_: str = "UndefinedObjectsAsExpectedValueTypes"
    code: int = 19
    description: str = (
        "Expected value types, which are stored in the column 'Type' in the 'Properties'"
        " sheet, are classes that exist in the 'Classes' sheet but for which no properties are defined "
        "in the 'Properties' sheet. "
    )
    example: str = (
        "We have 'Class1' which has property 'edgeClass1Class2' linking it to 'Class2', thus"
        "expected value of 'edgeClass1Class2' is 'Class2'. "
        "Both 'Class1' and 'Class2' are defined in the 'Classes' sheet"
        "However, only 'Class1' has properties defined in the 'Properties' sheet, making 'Class2' an undefined object"
        " leading to this error being raised!"
    )

    fix: str = (
        "Make sure to define properties for classes from 'Classes' "
        "sheet before defining properties that expect them as value types"
    )

    def __init__(self, undefined_objects: list[str], verbose: bool = False):
        self.message = (
            f"Following classes {', '.join(undefined_objects)} defined as classes in the 'Classes' sheet"
            f" have no properties defined in the 'Properties' sheet or their validation as objects failed!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ClassSheetParentClassIDRegexViolation(NeatException):
    """Parent ID, which is stored in the column 'Parent Class' in the 'Classes' sheet,
    does not respect defined regex expression

    Args:
        parent_ids: parent_ids that raised exception
        regex_expression: regex expression against which parent_id is validated
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Check definition of parent ids in `Parent Class` column in `Classes` sheet and
        make sure to respect the regex expression by removing any illegal characters

    """

    type_: str = "ClassSheetParentClassIDRegexViolation"
    code: int = 20
    description: str = (
        "Parent ID, which is stored in the column 'Parent Class' in the 'Classes' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If parent class is set to 'Class 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of class ids in 'Parent Class' column in 'Classes' sheet and "
        "make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, parent_ids: list[str], regex_expression: str, verbose: bool = False):
        self.parent_ids = parent_ids
        self.regex_expression = regex_expression

        self.message = (
            f"Parents ids: [{', '.join(parent_ids or [])}], stored in 'Parent Class' column in 'Classes' "
            f"sheet violates regex {self.regex_expression}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class MoreThanOneNonAlphanumericCharacter(NeatException):
    """This exceptions is raised when doing regex validation of strings which either
    represent class ids, property ids, prefix, data model name, that contain more than
    one non-alphanumeric character, such as for example '_' or '-'.

    Args:
        field_name: filed on which regex validation failed
        value: value that failed regex validation
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure not to use more than non-alphanumeric character in the row

    """

    type_: str = "MoreThanOneNonAlphanumericCharacter"
    code: int = 21
    description: str = (
        "This exceptions is raised when doing regex validation of strings which either"
        "represent class ids, property ids, prefix, data model name, that contain more than"
        "one non-alphanumeric character, such as for example '_' or '-'."
    )
    example: str = ""
    fix: str = ""

    def __init__(self, field_name: str, value: str, verbose: bool = False):
        self.field_name = field_name
        self.value = value

        self.message = (
            f"Field {field_name} with value {value} contains more than one non-alphanumeric character!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ViewExternalIdNotDefined(NeatException):
    """This exceptions is raised when external id of View is not defined.

    Args:
        field_name: filed on which regex validation failed
        value: value that failed regex validation
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "ViewExternalIdNotDefined"
    code: int = 22
    description: str = "This exceptions is raised when external id of View is not defined"
    example: str = ""
    fix: str = ""

    def __init__(self, verbose: bool = False):
        self.message = (
            f"Missing View external id!" f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class SpaceNotDefined(NeatException):
    """This exceptions is raised when CDF space name is missing.

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "SpaceNotDefined"
    code: int = 23
    description: str = "This exceptions is raised when CDF space name is missing"
    example: str = ""
    fix: str = ""

    def __init__(self, verbose: bool = False):
        self.message = (
            f"Missing CDF space name!" f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ViewVersionNotDefined(NeatException):
    """This exceptions is raised when View version is not provided.

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "ViewVersionNotDefined"
    code: int = 24
    description: str = "This exceptions is raised when View version is not provided"
    example: str = ""
    fix: str = ""

    def __init__(self, verbose: bool = False):
        self.message = (
            f"Missing View version!" f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class DefaultValueTypeNotProper(NeatException):
    """This exceptions is raised when default value type is not proper, i.e. it is not
    according to the expected value type set in Rules.


    Args:
        default_value_type: default value type that raised exception
        expected_value_type: expected value type that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "DefaultValueTypeNotProper"
    code: int = 25
    description: str = (
        "This exceptions is raised when default value type is not proper, i.e. it is not "
        "according to the expected value type set in Rules."
    )
    example: str = ""
    fix: str = ""

    def __init__(self, property_id: str, default_value_type: str, expected_value_type: str, verbose: bool = False):
        self.default_value_type = default_value_type
        self.expected_value_type = expected_value_type

        self.message = (
            f"Default value for property {property_id} is of type {default_value_type} which is different from "
            f"the expected value type {expected_value_type}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ClassToAssetMappingNotDefined(NeatException):
    """This exceptions is raised when deriving class to asset mapping when there is no
    mapping available.


    Args:
        class_id: Id of the class that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "ClassToAssetMappingNotDefined"
    code: int = 26
    description: str = (
        "This exceptions is raised when deriving class to asset mapping when there is no mapping available"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, class_id: str, verbose: bool = False):
        self.class_id = class_id

        self.message = (
            f"Requested serialization from pydantic model instance of class {class_id} is"
            " not possible since there is no class to asset mapping available!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PrefixAlreadyInUse(NeatException):
    """This exceptions is raised when trying to update base prefix/space of Rules object


    Args:
        class_id: Id of the class that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "PrefixAlreadyInUse"
    code: int = 27
    description: str = "This exceptions is raised when trying to update base prefix/space of Rules object"
    example: str = ""
    fix: str = ""

    def __init__(self, prefix: str, verbose: bool = False):
        self.prefix = prefix

        self.message = (
            f"Prefix {prefix} exist in self.prefixes, please use another prefix!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class IncompleteSchema(NeatException):
    """This exceptions is raised when schema is not complete, meaning defined properties
    are pointing to non-existing classes or value types


    Args:
        missing_classes: list of classes ids that are not defined in the sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    """

    type_: str = "IncompleteSchema"
    code: int = 28
    description: str = (
        "This exceptions is raised when schema is not complete, meaning "
        "defined properties are pointing to non-existing classes or value types"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_classes: set, verbose: bool = False):
        self.message = (
            f"Classes {missing_classes} are not defined in the Class sheet!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


# Warnings:


class ClassNameNotProvided(NeatWarning):
    """This warning is raised when class name is not provided in the 'Classes' sheet
    under 'name' column, which will be then set the class id stored in the 'Class' column

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
        class_id: class id that raised warning, and which is used as class name

    Notes:
        If you want to have differentiation between the class name and class id, then
        provide class name in the `name` column
    """

    type_: str = "ClassNameNotProvided"
    code: int = 1
    description: str = (
        "This warning is raised when class name is not provided in the 'Classes' sheet"
        "under 'name' column, which will be then set the class id stored in the 'Class' column"
    )
    example: str = ""
    fix: str = (
        "If you want to have differentiation between the class name"
        " and class id, then provide class name in the `name` column"
    )

    # need to have default value set to arguments
    # otherwise it will raise TypeError
    def __init__(self, class_id: str = "", verbose: bool = False):
        self.message = (
            f"Class id {class_id} set as Class name!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class EntityIDNotDMSCompliant(NeatWarning):
    """Warning raise when entity id being class, property or value type is not DMS compliant

    Args:
        entity_type: type of entity that raised warning
        entity_id: id of entity that raised warning
        loc: location of entity in the Transformation Rules sheet that raised warning
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        DMS ready means that entity id must only use following characters [a-zA-Z0-9_],
        where it can only start with letter! Also there are reserved words that cannot be used.

        Reserved words for views: `Query`, `Mutation`, `Subscription`, `String`, `Int32`, `Int64`, `Int`,
        `Float32`, `Float64`, `Float`, `Timestamp`, `JSONObject`, `Date`, `Numeric`, `Boolean`, `PageInfo`,
        `File`, `Sequence`, `TimeSeries`

        Reserved words for properties: `space`, `externalId`, `createdTime`, `lastUpdatedTime`,
        `deletedTime`, `edge_id`, `node_id`, `project_id`, `property_group`, `seq`, `tg_table_name`, `extensions`

        Reserved words for spaces: `space`, `cdf`, `dms`, `pg3`, `shared`, `system`, `node`, `edge`
    """

    type_: str = "EntityIDNotDMSCompliant"
    code: int = 2
    description: str = "Warning raise when entity id being class, property or value type is not DMS compliant"
    example: str = ""
    fix: str = (
        "DMS ready means that entity id must only use following"
        " characters [a-zA-Z0-9_], where it can only start with letter!"
    )

    # See ClassNameNotProvided for explanation why default values are set
    def __init__(self, entity_type: str = "", entity_id: str = "", loc: str = "", verbose: bool = False):
        self.message = (
            f"'{entity_id}' {entity_type.lower()}"
            " use character(s) outside of range of allowed characters [a-zA-Z0-9_] or "
            f"it starts with non-letter character or it is reserved word! {loc}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class PropertyRedefined(NeatWarning):
    """Warning raise when same property is defined multiple times for same class, this
    typically occurs if there are multiple ways to extract certain information from the
    NeatGraph.

    Args:
        property_id: property id that is redefined
        class_id: class id for which property is redefined
        loc: location of property redefinition in the 'Properties' sheet of Transformation Rules Excel file
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If possible, have only single definition of a particular property for a class, otherwise
        transformation rules will not be DMS compliant.
    """

    type_: str = "PropertyRedefined"
    code: int = 3
    description: str = "Warning raise when same property is defined multiple times for same class"
    example: str = ""
    fix: str = "Have only single definition of a particular property for a class"

    # See Warning302 for explanation why default values are set
    def __init__(self, property_id: str = "", class_id: str = "", loc: str = "", verbose: bool = False):
        self.message = (
            f"Not DMS compliant! Property '{property_id}' for class '{class_id}' redefined! {loc}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class PropertyNameNotProvided(NeatWarning):
    """If property name is not provided in the 'Property' sheet under 'name' column, it
    will be set to corresponding value from 'Property' column, thus property id

    Args:
        property_id: property id that is set as property name
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    Notes:
        If you want to have different property name then property id, provide it in the 'name' column
    """

    type_: str = "PropertyNameNotProvided"
    code: int = 4
    description: str = (
        "If property name is not provided in the 'Property' sheet under 'name' column,"
        " it will be set to corresponding value from 'Property' column, thus property id"
    )
    example: str = ""
    fix: str = "If you want to have different property name then property id, provide it in the 'name' column"

    def __init__(self, property_id: str = "", verbose: bool = False):
        self.message = (
            f"Property id {property_id} set as Property name!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class MissingLabel(NeatWarning):
    """If property maps to CDF relationship, and it does not have label explicitly stated
    under 'Label' column  in the 'Property' sheet under 'name' column,  it will be set
    to corresponding value from 'Property' column, thus property id

    Args:
        property_id: property id that which is missing label
        verbose: flag that indicates whether to provide enhanced exception message, by default False


    Notes:
        If you want to have control over relationship labels make sure to define one
        in the `Label` column in the `Properties` sheet
    """

    type_: str = "MissingLabel"
    code: int = 5
    description: str = (
        "If property maps to CDF relationship, and it does not have label explicitly stated under 'Label' column"
        " in the 'Property' sheet under 'name' column,"
        " it will be set to corresponding value from 'Property' column, thus property id"
    )
    example: str = ""
    fix: str = (
        "If you want to have control over relationship labels make sure to define one"
        " in the 'Label' column in the 'Properties' sheet."
    )

    def __init__(self, property_id: str = "", verbose: bool = False):
        self.message = (
            f"Property id {property_id} set as CDF relationship label!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class NoTransformationRules(NeatWarning):
    """This warning is raised if there are no transformation rules defined for given
    property and class

    Args:
        property_id: property id that which is missing transformation rules
        class_id: class id for which property is missing transformation rules
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        One can omit this warning if the Transformation Rules spreadsheet is used solely
        for defining the data model and not for performing knowledge graph transformation.
    """

    type_: str = "NoTransformationRules"
    code: int = 6
    description: str = (
        "This warning is raised if there are no transformation rules "
        "defined in the 'Transformation' sheet for given property"
    )
    example: str = ""
    fix: str = "No fix is provided for this warning"

    def __init__(self, property_id: str = "", class_id: str = "", verbose: bool = False):
        self.message = (
            f"There is no transformation rule configured for class '{class_id}' property '{property_id}'!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class NamespaceEndingFixed(NeatWarning):
    """This warning occurs when namespace does not end with '/' or '#'

    Args:
        namespace: namespace that raised warning due to improper ending
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure that namespace ends with `/` or `#`, if not it will be fixed by adding
        `#` at the end !
    """

    type_: str = "NamespaceEndingFixed"
    code: int = 7
    description: str = "This warning occurs when namespace does not end with '/' or '#'"
    example: str = "If namespace is set to http://purl.org/cognite, it will be converted to http://purl.org/cognite#"
    fix: str = "Make sure that namespace ends with '/' or '#'"

    def __init__(self, namespace: Namespace, verbose: bool = False):
        self.message = (
            f"Namespace {namespace} ending fixed by adding '#' at its end!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class DataModelIdMissing(NeatWarning):
    """This warning occurs when data model name is not provided in 'Metadata' sheet

    Args:
        prefix: prefix to which data model name will be set if not provided
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure that namespace ends with `/` or `#`, if not it will be fixed by adding
        `#` at the end !
    """

    type_: str = "DataModelIdMissing"
    code: int = 8
    description: str = "This warning occurs when data model id is not provided in 'Metadata' sheet"
    example: str = ""
    fix: str = (
        "Provide data model id by setting value for `external_id`,"
        " to avoid this warning and otherwise it will default to prefix"
    )

    def __init__(self, prefix: str, verbose: bool = False):
        self.message = (
            f"Data model id not provided, defaulting to prefix {prefix}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class VersionDotsConvertedToUnderscores(NeatWarning):
    """This warning occurs when converting version from dot notation to use of underscores
    in order to achieve version format that is accepted by CDF/DMS

    Args:
        prefix: prefix to which data model name will be set if not provided
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Typically version is expressed in classical form with dots major.minor.patch, while
        CDF accepts underscores major_minor_patch, thus this warning occurs when converting
        version from dot notation to use of underscores in order to achieve version format
        that is accepted by CDF/DMS
    """

    type_: str = "VersionDotsConvertedToUnderscores"
    code: int = 9
    description: str = (
        "This warning occurs when converting version from dot notation to use of underscores"
        " in order to achieve version format that is accepted by CDF/DMS"
    )
    example: str = "If version is provided as 1.2.3, this will be converted to 1_2_3 to be accepted by CDF"
    fix: str = "Convert version to underscore notation major_minor_patch"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Data model version expressed with '.' which is not acceptable for CDF."
            " All '.' are converted to '_'!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class DefaultValueTypeConverted(NeatWarning):
    """This exceptions is warning is raised when default value type is being converted to
    the expected value type set in Rules.


    Args:
        default_value_type: default value type that raised exception
        expected_value_type: expected value type that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure not to use more than non-alphanumeric character in the row

    """

    type_: str = "DefaultValueTypeConverted"
    code: int = 10
    description: str = (
        "This exceptions is warning is raised when default value type is being converted to "
        "the expected value type set in Rules."
    )
    example: str = ""
    fix: str = ""

    def __init__(
        self, property_id: str = "", default_value_type: str = "", expected_value_type: str = "", verbose: bool = False
    ):
        self.default_value_type = default_value_type
        self.expected_value_type = expected_value_type

        self.message = (
            f"Default value for property {property_id} is of type {default_value_type} "
            f"has been converted to the expected value type {expected_value_type}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class DefaultValueNotList(NeatWarning):
    """This exceptions is warning is raised when default value is not a list while it is
    expected to be due to set maximum cardinality being different than 1.


    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure not to use more than non-alphanumeric character in the row

    """

    type_: str = "DefaultValueNotList"
    code: int = 11
    description: str = (
        "This exceptions is warning is raised when default value is not a list while it is "
        "expected to be due to set maximum cardinality being different than 1"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, property_id: str, verbose: bool = False):
        self.message = (
            f"Default value for property {property_id} is not a list, "
            "while it is expected to be due to set maximum cardinality being different than 1!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


################################################################################################
# RULES PROCESSING: 100 - 199 ##################################################################
################################################################################################


class NotValidRDFPath(NeatException):
    """Provided `rdfpath` is not valid, i.e. it cannot be converted to SPARQL query.

    Args:
        rdf_path: `rdfpath` that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Get familiar with `rdfpath` to avoid this exception.
    """

    type_: str = "NotValidRDFPath"
    code: int = 100
    description: str = "Provided `rdfpath` is not valid, i.e. it cannot be converted to SPARQL query"
    example: str = ""
    fix: str = "Get familiar with `rdfpath` and check if provided path is valid!"

    def __init__(self, rdf_path: str, verbose: bool = False):
        self.rdf_path = rdf_path

        self.message = (
            f"{self.rdf_path} is not a valid rdfpath!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class NotValidTableLookUp(NeatException):
    """Provided `table lookup` is not valid, i.e. it cannot be converted to CDF lookup.

    Args:
        table_look_up: `table_look_up`, a part of `rawlookup`, that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Get familiar with `rawlookup` and `rdfpath` to avoid this exception.
    """

    type_: str = "NotValidTableLookUp"
    code: int = 101
    description: str = "Provided table lookup is not valid, i.e. it cannot be converted to CDF lookup"
    example: str = ""
    fix: str = "Get familiar with RAW look up and RDF paths and check if provided rawlookup is valid"

    def __init__(self, table_look_up: str, verbose: bool = False):
        self.table_look_up = table_look_up

        self.message = (
            f"{self.table_look_up} is not a valid table lookup"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class NotValidRAWLookUp(NeatException):
    """Provided `rawlookup` is not valid, i.e. it cannot be converted to SPARQL query and CDF lookup

    Args:
        raw_look_up: `rawlookup` rule that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Get familiar with `rawlookup` and `rdfpath` to avoid this exception.
    """

    type_: str = "NotValidRAWLookUp"
    code: int = 102
    description: str = "Provided rawlookup is not valid, i.e. it cannot be converted to SPARQL query and CDF lookup"
    example: str = ""
    fix: str = "Get familiar with `rawlookup` and `rdfpath` to avoid this exception"

    def __init__(self, raw_look_up: str, verbose: bool = False):
        self.raw_look_up = raw_look_up

        self.message = (
            f"Invalid rawlookup expected traversal | table lookup, got {raw_look_up}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


################################################################################################
# RULES IMPORTERS: 200 - 299 ###################################################################
################################################################################################
class RulesHasErrors(NeatWarning):
    """This warning occurs when generated transformation rules are invalid/incomplete.

    Args:
        importer_type: type of importer that is used to generate transformation rules
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Generated transformation rules are not guaranteed to be valid and complete.
        Go through the generated report file and fix the errors and warnings.
    """

    type_: str = "GeneratedTransformationRulesHasErrors"
    code: int = 200
    description: str = (
        "This warning occurs when transformation rules generated using an importer are invalid/incomplete."
    )
    example: str = ""
    fix: str = "Go through the generated report file and fix the warnings in generated Transformation Rules."

    def __init__(self, importer_type: str = "OWL ontology", verbose: bool = False):
        self.message = (
            f"Rules generated using {importer_type} are invalid!"
            " Consult generated validation report for details on the errors and fix them"
            " before using the rules file."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class RulesHasWarnings(NeatWarning):
    """This warning occurs when th generated transformation rules are invalid/incomplete.

    Args:
    importer_type: type of importer that is used to generate transformation rules
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        OWL ontology is not guaranteed to contain all information needed to generate
        transformation rules. In such cases, transformation rules generated from OWL ontology
        will be incomplete and invalid. Go through the generated report file and fix the warnings
    """

    type_: str = "GeneratedTransformationRulesHasWarnings"
    code: int = 201
    description: str = (
        "This warning occurs when transformation rules generated using an importer are invalid/incomplete."
    )
    example: str = ""
    fix: str = "Go through the generated report file and fix the warnings in generated Transformation Rules."

    def __init__(self, importer_type: str = "OWL ontology", verbose: bool = False):
        self.message = (
            f"Rules generated using {importer_type} raised warnings!"
            " Consult generated validation report for details on the warnings and optionally fix them"
            " before using the rules file."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class GraphClassNameCollision(NeatWarning):
    """This warning occurs when graph contains instances of classes with same name, but
    belonging to different namespaces.

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Since the RDF graph is fully flexible and based on URIs, it is possible to have
        instances of classes with same name, but belonging to different namespaces. This
        warning is raised when such collision occurs.
    """

    type_: str = "GraphClassNameCollision"
    code: int = 202
    description: str = (
        "This warning occurs when graph contains instances of classes with same name, but"
        " belonging to different namespaces."
    )
    example: str = ""
    fix: str = "Be caution when reviewing the generated transformation rules."

    def __init__(self, class_name: str, verbose: bool = False):
        self.message = f"Class name collision detected in the graph for class name {class_name}!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class GraphClassPropertyMultiValueTypes(NeatWarning):
    """This warning occurs when a same property is define for two object/classes where
    its expected value type is different in one definition, e.g. acts as an edge, while in
    other definition acts as and attribute

    Args:
        class_id: class id that raised warning due to multi type definition
        property_id: property id that raised warning due to multi type definition
        types: list of types of property
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If a property takes different value types for different objects, simply define
        new property. It is bad practice to have multi type property!
    """

    type_: str = "GraphClassPropertyMultiValueTypes"
    code: int = 203
    description: str = (
        "This warning occurs when a same property is define for two object/classes where"
        " its expected value type is different in one definition, e.g. acts as an edge, while in "
        "other definition acts as and attribute"
    )
    example: str = ""
    fix: str = "If a property takes different value types for different objects, simply define new property"

    def __init__(
        self, class_name: str = "", property_name: str = "", types: list[str] | None = None, verbose: bool = False
    ):
        self.message = (
            "It is bad practice to have multi type property! "
            f"Currently property '{property_name}' for class {class_name} has"
            f" multi type property: {', '.join(types or [])}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class GraphClassPropertyMultiOccurrence(NeatWarning):
    """This warning occurs when there is multiple different occurrences of the same property
    across various class instances.

    Args:
        class_id: class id that raised warning due to multi type definition
        property_id: property id that raised warning due to multi type definition
        occurrences: list of property occurrences
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If a property takes different value types for different objects, simply define
        new property. It is bad practice to have multi type property!
    """

    type_: str = "GraphClassPropertyMultiOccurrence"
    code: int = 204
    description: str = (
        "This warning occurs when there is multiple different occurrences of the same property "
        " across various class instances"
    )
    example: str = ""
    fix: str = "There"

    def __init__(self, class_name: str = "", property_name: str = "", verbose: bool = False):
        self.message = (
            f"Currently property '{property_name}' for class {class_name} has multi occurrences"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


################################################################################################
# RULES PARSERS: 300 - 399 #####################################################################
################################################################################################
class SourceObjectDoesNotProduceMandatorySheets(NeatException):
    """Given object (e.g., Excel file) does not produce one or more mandatory sheets

    Args:
        missing_mandatory_sheets: set of missing mandatory sheets
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        This exception is raised when source object converted to is missing one or more mandatory sheets.
        The mandatory sheets are: `Metadata`, `Classes`, `Properties`.
    """

    type_: str = "SourceObjectDoesProduceMandatorySheets"
    code: int = 300
    description: str = "Given Excel file is missing one or more mandatory sheets"
    example: str = "An Excel file is missing sheet named 'Metadata'"
    fix: str = "Make sure that Excel file contains all mandatory sheets, i.e. 'Metadata', 'Classes', 'Properties'"

    def __init__(self, missing_mandatory_sheets: set[str], verbose: bool = False):
        self.missing_fields = missing_mandatory_sheets

        self.message = (
            "Given Excel file is not compliant Transformation Rules file."
            f" It is missing mandatory sheets: {', '.join(missing_mandatory_sheets)}."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class UnableToDownloadExcelFile(NeatException):
    """This error is raised during loading of byte representation of  a Excel file from
    Github, when given file cannot be downloaded.

    Args:
        filepath: file path to Excel file that cannot be downloaded from Github
        loc: URL of Excel file from Github repository
        reason: reason why file cannot be downloaded (e.g., Forbidden, Not Found, etc.)
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure you provided correct parameters to download Excel file from github repository.
    """

    type_: str = "UnableToDownloadExcelFile"
    code: int = 301
    description: str = (
        "This error is raised during loading of byte representation of"
        " a Excel file from Github when given file cannot be downloaded."
    )
    example: str = ""
    fix: str = "Make sure you provided correct parameters to download Excel file from github repository"

    def __init__(self, filepath: str, loc: str, reason: str, verbose: bool = False):
        self.message = (
            f"File '{filepath}' from '{loc}' cannot be downloaded! Reason: {reason}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class NotExcelFile(NeatException):
    """This error is raised during loading of byte representation of a file from Github
    into `openpyxl` `Workbook` object in case when byte representation is not Excel file.

    Args:
        filepath: file path to Excel file that cannot be downloaded from Github
        loc: URL of Excel file from Github repository
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure you that byte representation of a file is Excel file.
    """

    type_: str = "NotExcelFile"
    code: int = 302
    description: str = (
        "This error is raised during loading of byte representation of a file from Github"
        " into `openpyxl` `Workbook` object in case when byte representation is not Excel file."
    )
    example: str = ""
    fix: str = "Make sure you that byte representation of a file is Excel file!"

    def __init__(self, filepath: str, loc: str, verbose: bool = False):
        self.message = (
            f"File '{filepath}' from '{loc}' is not a valid excel file!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MetadataSheetMissingMandatoryFields(NeatException):
    """Metadata sheet, which is part of Transformation Rules Excel file, is missing
    mandatory rows (i.e., fields)

    Args:
        missing_fields: Fields/rows that are missing in Metadata sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "MetadataSheetMissingMandatoryFields"
    code: int = 303
    description: str = "Metadata sheet, which is part of Transformation Rules Excel file, is missing mandatory rows"
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose: bool = False):
        self.missing_fields = missing_fields

        self.message = (
            f"Metadata sheet is missing following mandatory fields: {', '.join(missing_fields)}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ClassesSheetMissingMandatoryColumns(NeatException):
    """Classes sheet, which is a mandatory part of Transformation Rules Excel file, is
    missing mandatory columns at row 2

    Args:
        missing_fields: Fields/columns that are missing in Classes sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "ClassesSheetMissingMandatoryColumns"
    code: int = 304
    description: str = (
        "Classes sheet, which is a mandatory part of Transformation Rules Excel file, "
        "is missing mandatory columns at row 2"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose: bool = False):
        self.missing_fields = missing_fields

        self.message = (
            f"Classes sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 2"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertiesSheetMissingMandatoryColumns(NeatException):
    """Properties sheet, which is a mandatory part of Transformation Rules Excel file, is
    missing mandatory columns at row 2

    Args:
        missing_fields: Fields/columns that are missing in Properties sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "PropertiesSheetMissingMandatoryColumns"
    code: int = 305
    description: str = (
        "Properties sheet, which is a mandatory part of Transformation Rules Excel file, "
        "is missing mandatory columns at row 2"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose: bool = False):
        self.missing_fields = missing_fields

        self.message = (
            f"Properties sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 2"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PrefixesSheetMissingMandatoryColumns(NeatException):
    """Prefixes sheet, which is part of Transformation Rules Excel file, is missing
    mandatory columns at row 1

    Args:
        missing_fields: Fields/columns that are missing in Prefixes sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "PrefixesSheetMissingMandatoryColumns"
    code: int = 306
    description: str = (
        "Prefixes sheet, which is part of Transformation Rules Excel file, is missing mandatory columns at row 1"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose: bool = False):
        self.missing_fields = missing_fields

        self.message = (
            f"Prefixes sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 1"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class InstancesSheetMissingMandatoryColumns(NeatException):
    """Instances sheet, which is part of Transformation Rules Excel file, is missing
    mandatory columns at row 1

    Args:
        missing_fields: Fields/columns that are missing in Instances sheet
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "InstancesSheetMissingMandatoryColumns"
    code: int = 307
    description: str = (
        "Instances sheet, which is part of Transformation Rules Excel file, is missing mandatory columns at row 1"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose: bool = False):
        self.missing_fields = missing_fields

        self.message = (
            f"Instances sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 1"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


# Warnings
class MissingDataModelPrefixOrNamespace(NeatWarning):
    """Prefix and/or namespace are missing in the 'Metadata' sheet

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
    Add missing prefix and/or namespace in the 'Metadata' sheet
    """

    type_: str = "MissingDataModelPrefixOrNamespace"
    code: int = 300
    description: str = "Either prefix or namespace or both are missing in the 'Metadata' sheet"
    example: str = ""
    fix: str = "Add missing prefix and/or namespace in the 'Metadata' sheet"

    def __init__(self, verbose: bool = False):
        self.message = (
            "Instances sheet is present but prefix and/or namespace are missing in 'Metadata' sheet."
            "Instances sheet will not be processed!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


################################################################################################
# RULES EXPORTERS 400-499#######################################################################
################################################################################################


class EntitiesContainNonDMSCompliantCharacters(NeatException):
    """This error is raised during export of Transformation Rules to DMS schema when
    entities (e.g., types and fields) ids contain non DMS compliant characters.

    Args:
        report: report of entities that contain non DMS compliant characters
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure to check validation report of Transformation Rules and fix DMS related exceptions.
    """

    type_: str = "EntitiesContainNonDMSCompliantCharacters"
    code: int = 400
    description: str = (
        "This error is raised during export of Transformation Rules to"
        " DMS schema when entities contain non DMS compliant characters."
    )
    example: str = ""
    fix: str = "Make sure to check validation report of Transformation Rules and fix DMS related exceptions."

    def __init__(self, report: str = "", verbose: bool = False):
        self.message = (
            f"Following entities contain non DMS compliant characters: {report}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertiesDefinedMultipleTimes(NeatException):
    """This error is raised during export of Transformation Rules to DMS schema when
    when properties are defined multiple times for the same class.

    Args:
        report: report on properties which are defined multiple times
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
       Make sure to check validation report of Transformation Rules and fix DMS related warnings.
    """

    type_: str = "PropertiesDefinedMultipleTimes"
    code: int = 401
    description: str = (
        "This error is raised during export of Transformation Rules to "
        "DMS schema when properties are defined multiple times for the same class."
    )
    example: str = ""
    fix: str = "Make sure to check validation report of Transformation Rules and fix DMS related warnings."

    def __init__(self, report: str = "", verbose: bool = False):
        self.message = (
            f"Following properties defined multiple times for the same class(es): {report}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertyDefinitionsNotForSameProperty(NeatException):
    """This error is raised if property definitions are not for linked to the same
    property id when exporting rules to ontological representation.

    Args:
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "PropertyDefinitionsNotForSameProperty"
    code: int = 402
    description: str = "This error is raised if property definitions are not for linked to the same property id"
    example: str = ""
    fix: str = ""

    def __init__(self, verbose: bool = False):
        self.message = (
            "All definitions should have the same property_id! Aborting."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class FieldValueOfUnknownType(NeatException):
    """This error is raised when generating in-memory pydantic model from Transformation
    Rules from model, when field definitions are not provided as dictionary of field names
    ('str') and their types ('tuple' or 'dict').

    Args:
        field: field name that raised exception due to unknown type
        definition: definition of field that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "FieldValueOfUnknownType"
    code: int = 403
    description: str = (
        "This error is raised when generating in-memory pydantic model"
        " from Transformation Rules from model, when field definitions are not"
        " provided as dictionary of field names ('str') and their types ('tuple' or 'dict')."
    )
    example: str = ""
    fix: str = ""

    def __init__(self, field: str, definition: Any, verbose: bool = False):
        self.message = (
            f"Field {field} has definition of type {type(definition)}"
            " which is not acceptable! Only definition in form of dict or tuple is acceptable!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class MissingInstanceTriples(NeatException):
    """This error is raised when queried RDF class instance does not return any triples that define it.

    Args:
        id_: instance id
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "MissingInstanceTriples"
    code: int = 404
    description: str = (
        "This error is raised when queried RDF class instance " " does not return any triples that define it."
    )
    example: str = ""
    fix: str = "Make sure that RDF class instance holds necessary triples that define it."

    def __init__(self, id_: str | URIRef, verbose: bool = False):
        self.message = (
            f"Instance {id_} does not contain triples that would define it!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class PropertyRequiredButNotProvided(NeatException):
    """This error is raised when instantiating in-memory pydantic model from graph class
    instance which is missing required property.

    Args:
        id_: instance id
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "PropertyRequiredButNotProvided"
    code: int = 405
    description: str = (
        "This error is raised when instantiating in-memory pydantic model"
        " from graph class instance which is missing required property."
    )
    example: str = ""
    fix: str = "Either make field optional or add missing property to graph instance."

    def __init__(self, property: str, id_: str | URIRef, verbose: bool = False):
        self.message = (
            f"Property {property} is not present in graph instance {id_}!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class DataModelOrItsComponentsAlreadyExist(NeatException):
    """This error is raised when attempting to create data model which already exist in DMS

    Args:
        existing_data_model: external_id of model that already exist in DMS
        existing_containers: set of external_ids of containers that already exist in DMS
        existing_views: set of external_ids of views that already exist in DMS
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Remove existing data model and underlying views and/or containers, or bump
        version of data model and views and/ir optionally delete containers.
    """

    type_: str = "DataModelOrItsComponentsAlreadyExist"
    code: int = 406
    description: str = "This error is raised when attempting to create data model which already exist in DMS."
    example: str = ""
    fix: str = (
        "Remove existing data model and underlying views and/or containers, or bump "
        "version of data model and views and optionally delete containers."
    )

    def __init__(
        self,
        existing_spaces: set[str] | None,
        existing_data_model: DataModelId | None,
        existing_containers: set[str],
        existing_views: set[str],
        verbose: bool = False,
    ):
        self.existing_spaces = existing_spaces
        self.existing_data_model = existing_data_model
        self.existing_containers = existing_containers
        self.existing_views = existing_views

        self.message = "Aborting data model creation!"
        if self.existing_spaces:
            self.message += (
                f"\nSpaces {self.existing_spaces} already exists in DMS! Delete them or skip their creation! "
            )
        if self.existing_data_model:
            self.message += (
                f"\nData model {self.existing_data_model} already exists in DMS! Delete it first or bump its version! "
            )
        if self.existing_views:
            self.message += (
                f"\nViews {self.existing_views} already exist in DMS! Delete them first or bump their versions! "
            )
        if self.existing_containers:
            self.message += f"\nContainers {self.existing_containers} already exist in DMS! Delete them first! "

        self.message += (
            "\nTo remove existing data model and its components, use `self.remove_data_model(client)` method."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class InstancePropertiesNotMatchingViewProperties(NeatException):
    """This error is raised when an instance of a class has properties which are not
    defined in the DMS container

    Args:
        class_name: class name of instance that raised exception
        class_properties: list of mandatory properties of class
        container_properties: list of properties of container
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure that all properties of a class are defined in the DMS container.
    """

    type_: str = "InstancePropertiesNotMatchingViewProperties"
    code: int = 407
    description: str = "Instance of a class has properties which are not defined in the DMS view"
    example: str = ""
    fix: str = "Make sure that all properties of a class are defined in the DMS view"

    def __init__(self, class_name: str, class_properties: list[str], view_properties: list[str], verbose: bool = False):
        self.message = (
            f"Instance of class {class_name} has properties {class_properties}"
            f" while DMS view  {class_name} has properties {view_properties}!"
            f" Cannot create instance in DMS as properties do not match!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ContainerPropertyValueTypeRedefinition(NeatException):
    """This error is raised when building up container where a property being redefined
    with different value type

    Args:
        container_id: container id that raised exception
        property_id: container property id that raised exception
        value_type: value type of container property that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Make sure that when redefining property in container, value type is the same
    """

    type_: str = "ContainerPropertyValueTypeRedefinition"
    code: int = 408
    description: str = "Container property value type is being redefined"
    example: str = ""
    fix: str = "Make sure that when redefining property in container, value type remains the same"

    def __init__(
        self,
        container_id: str,
        property_id: str,
        current_value_type: str,
        redefined_value_type: str,
        loc: str,
        verbose: bool = False,
    ):
        self.message = (
            f"Container {container_id} property {property_id}"
            f" value type {current_value_type} redefined to {redefined_value_type}!"
            f"{loc}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class ViewPropertyRedefinition(NeatException):
    """This error is raised when building up views where a property being redefined differently

    Args:
        view_id: view id that raised exception
        property_id: view property id that raised exception
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Avoid redefining property in the same view
    """

    type_: str = "ViewPropertyRedefinition"
    code: int = 409
    description: str = "View property is being redefined in the same view but differently"
    example: str = ""
    fix: str = "Avoid redefining property in the same view"

    def __init__(
        self,
        view_id: str,
        property_id: str,
        loc: str,
        verbose: bool = False,
    ):
        self.message = (
            f"View {view_id} property {property_id} has been redefined in the same view!"
            f"{loc}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


# Warnings
class OntologyMultiTypeProperty(NeatWarning):
    """This warning occurs when a same property is define for two object/classes where
    its expected value type is different in one definition, e.g. acts as an edge, while in
    other definition acts as and attribute

    Args:
        property_id: property id that raised warning due to multi type definition
        types: list of types of property
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If a property takes different value types for different objects, simply define
        new property. It is bad practice to have multi type property!
    """

    type_: str = "OntologyMultiTypeProperty"
    code: int = 400
    description: str = (
        "This warning occurs when a same property is define for two object/classes where"
        " its expected value type is different in one definition, e.g. acts as an edge, while in "
        "other definition acts as and attribute"
    )
    example: str = ""
    fix: str = "If a property takes different value types for different objects, simply define new property"

    def __init__(self, property_id: str = "", types: list[str] | None = None, verbose: bool = False):
        self.message = (
            "It is bad practice to have multi type property! "
            f"Currently property '{property_id}' is defined as multi type property: {', '.join(types or [])}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class OntologyMultiRangeProperty(NeatWarning):
    """This warning occurs when a property takes range of values which consists of union
    of multiple value types

    Args:
        property_id: property id that raised warning due to multi range definition
        range_of_values: list of ranges that property takes
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If a property takes different range of values, simply define new property.
    """

    type_: str = "OntologyMultiRangeProperty"
    code: int = 401
    description: str = (
        "This warning occurs when a property takes range of values which consists of union of multiple value types."
    )
    example: str = ""
    fix: str = "If a property takes different range of values, simply define new property"

    def __init__(self, property_id: str = "", range_of_values: list[str] | None = None, verbose: bool = False):
        self.message = (
            "It is bad practice to have property that take various range of values! "
            f"Currently property '{property_id}' has multiple ranges: {', '.join(range_of_values or [])}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class OntologyMultiDomainProperty(NeatWarning):
    """This warning occurs when a property is reused for more than one classes

    Args:
        property_id: property id that raised warning due to reuse definition
        classes: list of classes that use the same property
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        No need to fix this, but make sure that property type is consistent across different
        classes and that ideally takes the same range of values
    """

    type_: str = "OntologyMultiDomainProperty"
    code: int = 402
    description: str = "This warning occurs when a property is reused for more than one classes."
    example: str = ""
    fix: str = (
        "No need to fix this, but make sure that property type is consistent"
        " across different classes and that ideally takes the same range of values"
    )

    def __init__(self, property_id: str = "", classes: list[str] | None = None, verbose: bool = False):
        self.message = (
            f"Currently property '{property_id}' is defined for multiple classes: {', '.join(classes or [])}"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class OntologyMultiLabeledProperty(NeatWarning):
    """This warning occurs when a property is given multiple labels, typically if the
    same property is defined for different classes but different name is given

    Args:
        property_id: property id that raised warning due to multiple labels
        names: list of names of property
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        This would be automatically fixed by taking the first label (aka name) of the property.
    """

    type_: str = "OntologyMultiLabeledProperty"
    code: int = 403
    description: str = (
        "This warning occurs when a property is given multiple labels,"
        " typically if the same property is defined for different "
        "classes but different name is given."
    )
    example: str = ""
    fix: str = "This would be automatically fixed by taking the first label (aka name) of the property."

    def __init__(self, property_id: str = "", names: list[str] | None = None, verbose: bool = False):
        self.message = (
            "Property should have single preferred label (human readable name)."
            f"Currently property '{property_id}' has multiple preferred labels: {', '.join(names or [])} !"
            f"Only the first name, i.e. '{names[0] if names else ''}' will be considered!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class OntologyMultiDefinitionProperty(NeatWarning):
    """This warning occurs when a property is given multiple human readable definitions,
    typically if the same property is defined for different classes where each definition
    is different.

    Args:
        property_id: property id that raised warning due to multiple definitions
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        This would be automatically fixed by concatenating all definitions.
    """

    type_: str = "OntologyMultiDefinitionProperty"
    code: int = 404
    description: str = (
        "This warning occurs when a property is given multiple human readable definitions,"
        " typically if the same property is defined for different "
        "classes where each definition is different."
    )
    example: str = ""
    fix: str = "This would be automatically fixed by concatenating all definitions."

    def __init__(self, property_id: str, verbose: bool = False):
        self.message = (
            f"Multiple definitions (aka comments) of property '{property_id}' detected."
            " Definitions will be concatenated."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class FieldNotFoundInInstance(NeatWarning):
    """This warning occurs when a property, associated to the pydantic field, is not found in the instance.
    The missing field will be removed, which might lead to failure of the pydantic model validation if
    the field/property is mandatory.

    Args:
        id_: instance id that raised warning due to missing field
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If property/field is mandatory make sure that instances contain all mandatory fields.
        Otherwise, no need to fix this warning.
    """

    type_: str = "FieldNotFoundInInstance"
    code: int = 405
    description: str = (
        "This warning occurs when a property, associated to the pydantic field, is not found in the instance."
        "The missing field will be removed, which might lead to failure of the pydantic model validation if"
        " the field/property is mandatory."
    )
    example: str = ""
    fix: str = (
        "If property/field is mandatory make sure that instances contain all mandatory fields."
        "Otherwise, no need to fix this warning."
    )

    def __init__(self, id_: str | URIRef = "", field_name: str = "", verbose: bool = False):
        self.message = (
            f"Field {field_name} is missing in the instance {id_}."
            " If this field is mandatory, the validation of the pydantic model will fail!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class FieldContainsMoreThanOneValue(NeatWarning):
    """This warning occurs when a property, associated to the pydantic field, contains
    more than one value (i.e. list of values), while it is defined as single value field.
    As consequence, only the first value will be considered!

    Args:
        field_name: field name that raised warning due to multiple values
        no_of_values: number of values that field contains
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If a property takes more than one value, define it as list of values in TransformationRules.
        To do this do not bound its `max_count` to 1, either leave it blank or set it to >1.
    """

    type_: str = "FieldContainsMoreThanOneValue"
    code: int = 406
    description: str = (
        "This warning occurs when a property, associated to the pydantic field, contains"
        " more than one value (i.e. list of values), while it is defined as single value field."
        " As consequence, only the first value will be considered!"
    )
    example: str = ""
    fix: str = (
        "If a property takes more than one value, define it as list of values in TransformationRules."
        "To do this do not bound its `max_count` to 1, either leave it blank or set it to >1."
    )

    def __init__(self, field_name: str = "", no_of_values: int | None = None, verbose: bool = False):
        self.message = (
            f"Field {field_name} is defined as single value property in TransformationRules,"
            f" but it contains {no_of_values} values!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class ContainerPropertyTypeUnsupported(NeatWarning):
    """This warning occurs when a property type is not supported by the container.
    Currently only `DatatypeProperty` and `ObjectProperty` are supported, which
    translate to `attribute` and `edge` respectively.

    Args:
        property_id: property id that raised warning due to unsupported type
        unsupported_type: unsupported property type
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Contact the NEAT support team if the warning is raised.
    """

    type_: str = "ContainerPropertyTypeUnsupported"
    code: int = 407
    description: str = (
        "This warning occurs when a property type is not supported by the container."
        " Currently only `DatatypeProperty` and `ObjectProperty` are supported, which"
        " translate to `attribute` and `edge` respectively."
    )
    example: str = ""
    fix: str = "Contact NEAT support team."

    def __init__(self, property_id: str = "", unsupported_type: str = "", verbose: bool = False):
        self.message = (
            f"Property {property_id} has unsupported type {unsupported_type}!"
            "Only DatatypeProperty and ObjectProperty are supported!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class ViewPropertyTypeUnsupported(NeatWarning):
    """This warning occurs when a TransformationRule property translates to unsupported
    DMS view property. Currently only attributes, edges 1-1 and edges 1-n are supported.

    Args:
        property_id: property id that raised warning due to unsupported type
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Contact the NEAT support team if the warning is raised.
    """

    type_: str = "ViewPropertyTypeUnsupported"
    code: int = 408
    description: str = (
        "This warning occurs when a TransformationRule property translates to unsupported DMS view property."
        " Currently only attributes, edges 1-1 and edges 1-n are supported."
    )
    example: str = ""
    fix: str = "Contact NEAT support team."

    def __init__(self, property_id: str = "", verbose: bool = False):
        self.message = (
            f"Property {property_id} translates to unsupported!"
            " Currently only attributes, edges 1-1 and edges 1-n are supported."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class ContainersAlreadyExist(NeatWarning):
    """This warning occurs when attempting to create containers which already exist in DMS.

    Args:
        container_ids: set of container ids that already exist in DMS
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If you need to create containers remove existing once and try again.
    """

    type_: str = "ContainersAlreadyExist"
    code: int = 409
    description: str = "This warning occurs when attempting to create containers which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing containers and try again."

    def __init__(self, container_ids: set[ContainerId] | None = None, space: str = "", verbose: bool = False):
        self.message = (
            f"Containers {container_ids or set()} already exist in space {space}. "
            "Since update of containers can cause issues, "
            "remove them first prior data model creation!"
            "Aborting containers creation!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class ViewsAlreadyExist(NeatWarning):
    """This warning occurs when attempting to create views which already exist in DMS.

    Args:
        views_ids: set of view ids that already exist in DMS
        version: version of data model/views that already exist in DMS
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        If you need to create views remove existing once and try again or update
        data model version.
    """

    type_: str = "ViewsAlreadyExist"
    code: int = 410
    description: str = "This warning occurs when attempting to create views which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing views and try again or update version of data model."

    def __init__(self, views_ids: set[ViewId] | None = None, version: str = "", space: str = "", verbose: bool = False):
        self.message = (
            f"Views {views_ids or set()} version {version} already exist in space {space}. "
            "Since update of views raise issues, "
            "remove them first prior data model creation or update version of data model!"
            "Aborting views creation!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class DataModelAlreadyExist(NeatWarning):
    """This warning occurs when attempting to create data model which already exist in DMS.

    Args:
        data_model_id: data model id that already exist in DMS
        version: version of data model that already exist in DMS
        verbose: flag that indicates whether to provide enhanced exception message, by default False

    Notes:
        Remove existing data model and try again or update its version
    """

    type_: str = "DataModelAlreadyExist"
    code: int = 411
    description: str = "This warning occurs when attempting to create data model which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing data model and try again or update its version."

    def __init__(self, data_model_id: str = "", version: str = "", space: str = "", verbose: bool = False):
        self.message = (
            f"Data model {data_model_id} version {version} already exist in space {space}. "
            "Since update of data model can raise issues, "
            "remove it first or update its version!"
            "Aborting data model creation!"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class EdgeConditionUnmet(NeatWarning):
    """This warning occurs when attempting to create an edge but not all conditions are met.

    Args:
        edge: data model id that already exist in DMS
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "EdgeConditionUnmet"
    code: int = 412
    description: str = "This warning occurs when attempting to create an edge, but the conditions are not met."
    example: str = ""
    fix: str = "Check if the edge is valid and that the lenght of the external_id is < 255"

    def __init__(self, edge: str, verbose: bool = False):
        self.message = (
            f"Ignoring edge {edge} as its format is not valid"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"



from cognite.neat.issues import MultiValueError

from . import dms, fileread, formatters, importing, spreadsheet, spreadsheet_file
from .base import (
    DefaultPydanticError,
    IssueList,
    NeatValidationError,
    ValidationIssue,
    ValidationWarning,
)

__all__ = [
    "DefaultPydanticError",
    "MultiValueError",
    "IssueList",
    "NeatValidationError",
    "ValidationIssue",
    "ValidationIssue",
    "ValidationWarning",
    "dms",
    "fileread",
    "formatters",
    "importing",
    "spreadsheet",
    "spreadsheet_file",
]
